from schematics.models import Model
from schematics.types import StringType, ListType


class CLICommandExample(Model):
    name = StringType(required=True)
    commands = ListType(StringType(), required=True, min_size=1)

    class Options:
        serialize_when_none = False



from .reference import ReferenceField


class XmsODataField(ReferenceField):
    """
    When present the x-ms-odata extensions indicates the operation includes one or more OData query parameters. These parameters include $filter, $top, $orderby, $skip, and $expand. In some languages the generated method will expose these parameters as strongly types OData type.

    Schema: ref to the definition that describes object used in filter.
    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-odata
    """

    def __init__(self, **kwargs):
        super(XmsODataField, self).__init__(
            serialized_name='x-ms-odata',
            deserialize_from='x-ms-odata',
            **kwargs
        )



from schematics.models import Model
from schematics.types import BaseType, StringType, FloatType, ModelType, BooleanType, IntType, ListType

from command.model.configuration import CMDArraySchemaBase
from .fields import DataTypeFormatEnum, RegularExpressionField, XNullableField
from .x_ms_enum import XmsEnumField


class Items(Model):
    """A limited subset of JSON-Schema's items object. It is used by parameter definitions that are not located in "body"."""

    type = StringType(
        choices=("string", "number", "integer", "boolean", "array"),
        required=True
    )  # Required. The type of the object. The value MUST be one of "string", "number", "integer", "boolean", or "array".

    format = DataTypeFormatEnum()  # The extending format for the previously mentioned type. See Data Type Formats for further details.

    collection_format = StringType(
        choices=("csv", "ssv", "tsv", "pipes"),  # default is csv
        serialized_name="collectionFormat",
        deserialize_from="collectionFormat",
    )  # Determines the format of the array if type array is used. Possible values are: csv - comma separated values foo,bar; ssv - space separated values foo bar; tsv - tab separated values foo\tbar; pipes - pipe separated values foo|bar.

    default = BaseType()  # This keyword can be used to supply a default JSON value associated with a particular schema.  It is RECOMMENDED that a default value be valid against the associated schema.

    enum = ListType(BaseType())

    # Validation keywords for numeric instances (number and integer)
    multiple_of = FloatType(
        min_value=0,
        serialized_name="multipleOf",
        deserialize_from="multipleOf",
    )  # The value of "multipleOf" MUST be a JSON number.  This number MUST be strictly greater than 0.
    maximum = FloatType()
    exclusive_maximum = BooleanType(
        serialized_name="exclusiveMaximum",
        deserialize_from="exclusiveMaximum",
    )
    minimum = FloatType()
    exclusive_minimum = BooleanType(
        serialized_name="exclusiveMinimum",
        deserialize_from="exclusiveMinimum"
    )

    # Validation keywords for strings
    max_length = IntType(
        min_value=0,
        serialized_name="maxLength",
        deserialize_from="maxLength"
    )
    min_length = IntType(
        min_value=0,
        serialized_name="minLength",
        deserialize_from="minLength"
    )
    pattern = RegularExpressionField()

    # Validation keywords for arrays
    items = ModelType("Items")  # Required if type is "array". Describes the type of items in the array.
    max_items = IntType(
        min_value=0,
        serialized_name="maxItems",
        deserialize_from="maxItems"
    )
    min_items = IntType(
        min_value=0,
        serialized_name="minItems",
        deserialize_from="minItems"
    )
    unique_items = BooleanType(
        serialized_name="uniqueItems",
        deserialize_from="uniqueItems"
    )

    x_ms_enum = XmsEnumField()
    x_nullable = XNullableField()  # TODO: # when true, specifies that null is a valid value for the associated schema

    def to_cmd(self, builder, **kwargs):
        model = builder.build_schema(self)
        builder.setup_fmt(model, self)
        builder.setup_enum(model, self)
        builder.setup_default(model, self)
        if isinstance(model, CMDArraySchemaBase):
            if self.items:
                assert isinstance(self.items, Items)
                model.item = builder(self.items, in_base=True)
        return model



import click
import logging
from flask import Blueprint
import os
import json

from utils.config import Config

logger = logging.getLogger('aaz')

bp = Blueprint('swagger-cmds', __name__, url_prefix='/Swagger/CMDs', cli_group="swagger")
bp.cli.short_help = "Manage azure-rest-api-specs/azure-rest-api-specs-pr repos."

@bp.cli.command("export-resources", short_help="Export all control plane resources in swagger repo")
@click.option(
    "--swagger-path", '-s',
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_PATH,
    required=not Config.SWAGGER_PATH,
    callback=Config.validate_and_setup_swagger_path,
    expose_value=False,
    help="The local path of azure-rest-api-specs repo. Official repo is https://github.com/Azure/azure-rest-api-specs"
)
@click.option(
    "--output-file", '-o',
    type=click.Path(file_okay=True, dir_okay=False, resolve_path=True),
    help="The file name for result output",
)
def export_swagger_control_plane_resources(output_file):
    logger.setLevel(level=logging.CRITICAL)
    from swagger.controller.specs_manager import SwaggerSpecsManager
    from utils.plane import PlaneEnum

    if output_file and os.path.exists(output_file):
        print(f"The output file {output_file} exists. Do you want to override it? (Y/N):", end='')
        if input().lower() != 'y':
            print("Exit without overriding.")
            return

    result = {}
    swagger_specs = SwaggerSpecsManager()
    for module in swagger_specs.get_modules(PlaneEnum.Mgmt):
        for rp in module.get_resource_providers():
            if rp.name.lower() not in result:
                result[rp.name.lower()] = {}
            resource_map = result[rp.name.lower()]
            for resource_id, version_map in rp.get_resource_map().items():
                if resource_id not in resource_map:
                    resource_map[resource_id] = set()
                resource_map[resource_id].update([version for version in version_map])
    for rp, resource_map in result.items():
        for resource in resource_map:
            assert None not in resource_map[resource]
            resource_map[resource] = sorted(resource_map[resource])

    if output_file is None:
        print(json.dumps(result, ensure_ascii=True, sort_keys=True, indent=2))
    else:
        with open(output_file, 'w') as f:
            json.dump(result, f, ensure_ascii=True, sort_keys=True, indent=2)
        print(f'Export output in {output_file} success.')





def create_data_plane_name(resource_provider):
    return 'data-plane:' + resource_provider.lower()


class PlaneEnum:
    Mgmt = "mgmt-plane"
    _Data = "data-plane"

    @staticmethod
    def Data(resource_provider):
        """For data plane, it's required to provide resource_provider which defined the scope of data plane."""
        return create_data_plane_name(resource_provider)

    _config = {
        Mgmt: {
            "displayName": "Control plane",
            "client": "MgmtClient",  # MgmtClient is implemented in azure.cli.core.aaz._client
        },
        # the general data-plane configuration
        _Data: {
            "displayName": "Data plane",
            # DataPlaneClient is `NOT` implemented, the client configuration with endpoints and scopes is required.
            "client": "DataPlaneClient",
        },
        # example for special data plane configuration
        # create_data_plane_name("Microsoft.Insights"): {
        #     "displayName": "Data plane for Microsoft.Insights",
        #     "client": "InsightDataPlaneClient",
        #     "swagger_modules": ["monitor"],
        # }
    }

    @classmethod
    def is_valid_swagger_module(cls, plane, module_name):
        swagger_modules = cls._config.get(plane, {}).get('swagger_modules', None)
        if not swagger_modules or module_name in swagger_modules:
            return True
        return False
    
    @classmethod
    def is_data_plane(cls, plane):
        return plane.split(":")[0] == cls._Data
    
    @classmethod
    def get_data_plane_scope(cls, plane):
        if not cls.is_data_plane(plane) or plane == cls._Data:
            return None
        return plane.replace(cls._Data + ':', '') or None

    @classmethod
    def http_client(cls, plane):
        if plane in cls._config:
            return cls._config[plane]['client']
        elif cls.is_data_plane(plane):
            # use the default data plane client
            return cls._config[cls._Data]['client']
        else:
            raise ValueError(f"Invalid plane name '{plane}'")


__all__ = ["PlaneEnum"]



from ._command import CLIAtomicCommand, CLIAtomicCommandRegisterInfo
from ._command_group import CLIAtomicCommandGroup, CLIAtomicCommandGroupRegisterInfo
from ._example import CLICommandExample
from ._help import CLICommandHelp, CLICommandGroupHelp
from ._profile import CLIAtomicProfile
from ._resource import CLISpecsResource
from ._client import CLIAtomicClient



import json

from command.model.configuration import CMDClientConfig


class ClientCfgReader:

    def __init__(self, cfg):
        assert isinstance(cfg, CMDClientConfig)
        self.cfg = cfg

    def find_arg_by_var(self, arg_var):
        if not self.cfg.arg_group:
            return None
        for arg in self.cfg.arg_group.args:
            if arg.var == arg_var:
                return arg
        return None

    def get_cfg_file_data(self):
        data = json.dumps(self.cfg.to_primitive(), ensure_ascii=False)
        return data



from schematics.models import Model
from schematics.types import ModelType, ListType, DictType
from .operation import Operation
from .parameter import ParameterField, ParameterBase
from .reference import Linkable, Reference


class PathItem(Model, Linkable):
    """Describes the operations available on a single path. A Path Item may be empty, due to ACL constraints. The path itself is still exposed to the documentation viewer but they will not know which operations and parameters are available"""

    get = ModelType(Operation)  # A definition of a GET operation on this path.
    post = ModelType(Operation)  # A definition of a POST operation on this path.
    put = ModelType(Operation)  # A definition of a PUT operation on this path.
    patch = ModelType(Operation)  # A definition of a PATCH operation on this path.
    delete = ModelType(Operation)  # A definition of a DELETE operation on this path.
    head = ModelType(Operation)  # A definition of a HEAD operation on this path.

    parameters = ListType(ParameterField(support_reference=True))  # A list of parameters that are applicable for all the operations described under this path. These parameters can be overridden at the operation level, but cannot be removed there. The list MUST NOT include duplicated parameters. A unique parameter is defined by a combination of a name and location. The list can use the Reference Object to link to parameters that are defined at the Swagger Object's parameters. There can be one "body" parameter at most.

    # ignored because it is not used
    # options = ModelType(Operation)  # A definition of a OPTIONS operation on this path.
    # ref = ReferenceType()  # Allows for an external definition of this path item. The referenced structure MUST be in the format of a Path Item Object. If there are conflicts between the referenced definition and this Path Item's definition, the behavior is undefined.

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        if self.parameters is not None:
            for idx, param in enumerate(self.parameters):
                if isinstance(param, Linkable):
                    param.link(swagger_loader, *self.traces, 'parameters', idx)

            # replace parameter reference by parameter instance
            for idx in range(len(self.parameters)):
                param = self.parameters[idx]
                while isinstance(param, Reference):
                    param = param.ref_instance
                assert isinstance(param, ParameterBase)
                self.parameters[idx] = param

        if self.get is not None:
            self.get.link(swagger_loader, *self.traces, 'get')
        if self.put is not None:
            self.put.link(swagger_loader, *self.traces, 'put')
        if self.post is not None:
            self.post.link(swagger_loader, *self.traces, 'post')
        if self.delete is not None:
            self.delete.link(swagger_loader, *self.traces, 'delete')
        if self.head is not None:
            self.head.link(swagger_loader, *self.traces, 'head')
        if self.patch is not None:
            self.patch.link(swagger_loader, *self.traces, 'patch')

    def to_cmd(self, builder, **kwargs):
        op = getattr(self, builder.method, None)
        if op is None:
            return None
        assert isinstance(op, Operation)
        host_path, parent_parameters = builder.parse_parameterized_host_path()
        parent_parameters = parent_parameters or []
        if self.parameters:
            parent_parameters.extend(self.parameters)
        cmd_op = builder(op, parent_parameters=parent_parameters, host_path=host_path)
        return cmd_op


class PathsField(DictType):

    def __init__(self, **kwargs):
        super(PathsField, self).__init__(
            ModelType(PathItem),
            **kwargs
        )


class XmsPathsField(DictType):

    """
    OpenAPI 2.0 has a built-in limitation on paths. Only one operation can be mapped to a path and http method. There are some APIs, however, where multiple distinct operations are mapped to the same path and same http method. For example GET /mypath/query-drive?op=file and GET /mypath/query-drive?op=folder may return two different model types (stream in the first example and JSON model representing Folder in the second). Since OpenAPI does not treat query parameters as part of the path the above 2 operations may not co-exist in the standard "paths" element.

    To overcome this limitation an "x-ms-paths" extension was introduced parallel to "paths". URLs under "x-ms-paths" are allowed to have query parameters for disambiguation, however they are not actually used.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-paths
    """

    def __init__(self, **kwargs):
        super(XmsPathsField, self).__init__(
            ModelType(PathItem),
            serialized_name="x-ms-paths",
            deserialize_from="x-ms-paths",
            **kwargs
        )




from schematics.models import Model
from schematics.types import StringType, ListType, ModelType, PolyModelType
from schematics.types.serializable import serializable

from ._fields import CMDStageField, CMDVariantField, CMDPrimitiveField, CMDBooleanField, CMDClassField
from ._format import CMDStringFormat, CMDIntegerFormat, CMDFloatFormat, CMDObjectFormat, CMDArrayFormat, \
    CMDResourceIdFormat, CMDDateTimeFormat
from ._help import CMDArgumentHelp
from utils import exceptions

import copy


class CMDArgEnumItem(Model):
    # properties as tags
    name = StringType(required=True)
    hide = CMDBooleanField()

    # properties as nodes
    value = CMDPrimitiveField()  # json value format string, support null

    class Options:
        serialize_when_none = False

    @classmethod
    def build_enum_item(cls, schema_item, ref_enum_item):
        item = cls()
        item.value = copy.deepcopy(schema_item.value)
        item.hide = ref_enum_item.hide if ref_enum_item else None
        item.name = ref_enum_item.name if ref_enum_item else str(item.value)
        return item


class CMDArgEnum(Model):
    # properties as tags

    # properties as nodes
    items = ListType(ModelType(CMDArgEnumItem), min_size=1)
    support_extension = CMDBooleanField(
        serialized_name="supportExtension",
        deserialize_from="supportExtension",
    )

    @classmethod
    def build_enum(cls, schema_enum, ref_enum):
        enum = cls()
        enum.items = []
        for schema_item in schema_enum.items:
            ref_enum_item = None
            if ref_enum:
                for enum_item in ref_enum.items:
                    if enum_item.value == schema_item.value:
                        ref_enum_item = enum_item
                        break
            item = CMDArgEnumItem.build_enum_item(schema_item, ref_enum_item)
            enum.items.append(item)
        if ref_enum:
            enum.support_extension = ref_enum.support_extension
        return enum

    def reformat(self, **kwargs):
        self.items = sorted(self.items, key=lambda it: it.name)


class CMDArgDefault(Model):
    """ The argument value if an argument is not used """

    # properties as nodes
    value = CMDPrimitiveField()  # json value format string, support null

    @classmethod
    def build_default(cls, builder, schema_default):
        default = cls()
        default.value = copy.deepcopy(schema_default.value)
        return default


class CMDArgBlank(Model):
    """ The argument value if an argument is assigned empty

    use cases:

    boolean type:
        If it is blank, it will be assigned true.

    int, float type:
        If it is not required, then it will be assigned null, which means to remove this property in request.

    string type:
        It should be distinguish with empty string.
        If it is not required, then it will be assigned null, which means to remove this property in request.

    object type:
        If it is required, then it will be assigned an empty dict
        If it is not required, then it will be assigned null, which means to remove this property in request.
    """

    # properties as nodes
    value = CMDPrimitiveField()  # json value format string, support null


class CMDArgPromptInput(Model):
    """ Support the prompt input for """

    # The prompt string, if given, is printed to standard output without a trailing newline before reading input.
    msg = StringType(required=True)


class CMDPasswordArgPromptInput(CMDArgPromptInput):

    # Ask the user to re-enter the same value to confirm. It's supported for CMDPasswordArg only.
    confirm = CMDBooleanField()


class CMDArgBase(Model):
    TYPE_VALUE = None

    # base types: "array", "boolean", "integer", "float", "object", "string",
    # special types: "ResourceId", "ResourceGroupName", "SubscriptionId", "ResourceLocation", "File"

    nullable = CMDBooleanField()  # whether can pass null value or not.

    blank = ModelType(CMDArgBlank)  # blank value is used when argument don't have any value

    class Options:
        serialize_when_none = False

    @serializable
    def type(self):
        return self._get_type()

    def _get_type(self):
        assert self.TYPE_VALUE is not None
        return self.TYPE_VALUE

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.TYPE_VALUE is None:
            return False

        if isinstance(data, dict):
            type_value = data.get('type', None)
            if type_value is not None:
                typ = type_value.replace("<", " ").replace(">", " ").strip().split()[0]
                return typ == cls.TYPE_VALUE
        elif isinstance(data, CMDArgBase):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False

    @classmethod
    def build_arg_base(cls, builder):
        arg_base = cls()
        arg_base.nullable = builder.get_nullable()
        arg_base.blank = builder.get_blank()
        return arg_base

    def _reformat_base(self, **kwargs):
        pass

    def reformat(self, **kwargs):
        self._reformat_base(**kwargs)


class CMDArgBaseField(PolyModelType):

    def __init__(self, **kwargs):
        super(CMDArgBaseField, self).__init__(
            model_spec=CMDArgBase,
            allow_subclasses=True,
            serialize_when_none=False,
            **kwargs
        )

    def find_model(self, data):
        if self.claim_function:
            kls = self.claim_function(self, data)
            if not kls:
                raise Exception("Input for polymorphic field did not match any model")
            return kls

        fallback = None
        matching_classes = set()
        for kls in self._get_candidates():
            if issubclass(kls, CMDArg):
                continue

            try:
                kls_claim = kls._claim_polymorphic
            except AttributeError:
                if not fallback:
                    fallback = kls
            else:
                if kls_claim(data):
                    matching_classes.add(kls)

        if not matching_classes and fallback:
            return fallback
        elif len(matching_classes) != 1:
            raise Exception("Got ambiguous input for polymorphic field")

        return matching_classes.pop()


class CMDArg(CMDArgBase):
    # properties as tags
    var = CMDVariantField(required=True)
    options = ListType(StringType(), min_size=1, required=True)  # argument option names, the name is in dash case "aa-bb-cc"
    required = CMDBooleanField()
    stage = CMDStageField()

    hide = CMDBooleanField()
    group = StringType()   # argument group name
    id_part = StringType(
        serialized_name='idPart',
        deserialize_from='idPart',
    )   # for Resource Id argument

    # properties as nodes
    help = ModelType(CMDArgumentHelp)
    default = ModelType(CMDArgDefault)  # default value is used when argument isn't in command
    configuration_key = StringType(
        serialized_name='configurationKey',
        deserialize_from='configurationKey',
    )  # the key to retrieve the default value from cli configuration
    prompt = ModelType(CMDArgPromptInput)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_schema = None

    @classmethod
    def _claim_polymorphic(cls, data):
        if super()._claim_polymorphic(data):
            if isinstance(data, dict):
                # distinguish with CMDArgBase and CMDArg
                return 'var' in data
            else:
                return isinstance(data, CMDArg)
        return False

    @classmethod
    def build_arg(cls, builder):
        arg = cls.build_arg_base(builder)
        assert isinstance(arg, CMDArg)
        arg.var = builder.get_var()
        arg.options = builder.get_options()
        arg.help = builder.get_help()
        arg.group = builder.get_group()

        arg.required = builder.get_required()
        arg.default = builder.get_default()
        arg.configuration_key = builder.get_configuration_key()
        arg.prompt = builder.get_prompt()
        arg.hide = builder.get_hide()
        return arg

    def _reformat(self, **kwargs):
        self.options = sorted(self.options, key=lambda op: (len(op), op))

    def reformat(self, **kwargs):
        self._reformat_base(**kwargs)
        self._reformat(**kwargs)


#cls
class CMDClsArgBase(CMDArgBase):
    _type = StringType(
        deserialize_from='type',
        serialized_name='type',
        required=True
    )

    def _get_type(self):
        return self._type

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.implement = None

    @classmethod
    def _claim_polymorphic(cls, data):
        if isinstance(data, dict):
            type_value = data.get('type', None)
            if type_value is not None and type_value.startswith("@"):
                return True
        elif isinstance(data, CMDClsArgBase):
            return True
        return False

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        arg._type = builder.get_type()
        return arg

    def get_unwrapped(self, **kwargs):
        if self.implement is None:
            return None
        assert isinstance(self.implement, CMDArgBase)
        if isinstance(self.implement, CMDObjectArgBase):
            cls = CMDObjectArg if isinstance(self, CMDClsArg) else CMDObjectArgBase
        elif isinstance(self.implement, CMDArrayArgBase):
            cls = CMDArrayArg if isinstance(self, CMDClsArg) else CMDArrayArgBase
        else:
            raise NotImplementedError()
        data = {k: v for k, v in self.implement.to_native().items() if k in cls._schema.valid_input_keys}
        data.update(kwargs)
        unwrapped = cls(data)
        return unwrapped


class CMDClsArg(CMDClsArgBase, CMDArg):
    singular_options = ListType(
        StringType(),
        serialized_name='singularOptions',
        deserialize_from='singularOptions',
    )  # for list use only

    @classmethod
    def build_arg(cls, builder):
        arg = super().build_arg(builder)
        assert isinstance(arg, CMDClsArg)
        arg.singular_options = builder.get_singular_options()
        return arg

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.singular_options:
            self.singular_options = sorted(self.singular_options, key=lambda op: (len(op), op))

    def get_unwrapped(self, **kwargs):
        uninherent = {
            "var": self.var,
            "options": [*self.options],
            "required": self.required,
            "stage": self.stage,
            "hide": self.hide,
            "group": self.group,
            "id_part": self.id_part,
            "help": self.help.to_native() if self.help else None,
            "default": self.default.to_native() if self.default else None,
        }
        uninherent.update(kwargs)
        return super().get_unwrapped(**uninherent)


# string
class CMDStringArgBase(CMDArgBase):
    TYPE_VALUE = "string"

    fmt = ModelType(
        CMDStringFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    enum = ModelType(CMDArgEnum)

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDStringArgBase)
        arg.fmt = builder.get_fmt()
        arg.enum = builder.get_enum()
        return arg

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.enum:
            self.enum.reformat(**kwargs)
        if self.blank:
            if not isinstance(self.blank.value, str) and not (self.nullable and self.blank.value is None):
                raise exceptions.VerificationError(
                    f"Invalid blank value", details=f"'{self.blank.value}' is not a valid string arg value")


class CMDStringArg(CMDStringArgBase, CMDArg):

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.default:
            if not isinstance(self.default.value, str) and not (self.nullable and self.default.value is None):
                raise exceptions.VerificationError(
                    f"Invalid default value", details=f"'{self.default.value}' is not a valid string arg value")


# byte: base64 encoded characters
class CMDByteArgBase(CMDStringArgBase):
    TYPE_VALUE = "byte"


class CMDByteArg(CMDByteArgBase, CMDStringArg):
    pass


# binary: any sequence of octets
class CMDBinaryArgBase(CMDStringArgBase):
    TYPE_VALUE = "binary"


class CMDBinaryArg(CMDBinaryArgBase, CMDStringArg):
    pass


# duration
class CMDDurationArgBase(CMDStringArgBase):
    TYPE_VALUE = "duration"


class CMDDurationArg(CMDDurationArgBase, CMDStringArg):
    pass


# date: As defined by full-date - https://xml2rfc.tools.ietf.org/public/rfc/html/rfc3339.html#anchor14
class CMDDateArgBase(CMDStringArgBase):
    TYPE_VALUE = "date"


class CMDDateArg(CMDDateArgBase, CMDStringArg):
    pass


# date-time: As defined by date-time - https://xml2rfc.tools.ietf.org/public/rfc/html/rfc3339.html#anchor14
class CMDDateTimeArgBase(CMDStringArgBase):
    TYPE_VALUE = "dateTime"

    fmt = ModelType(
        CMDDateTimeFormat,
        serialized_name='format',
        deserialize_from='format'
    )


class CMDDateTimeArg(CMDDateTimeArgBase, CMDStringArg):
    pass


class CMDTimeArgBase(CMDStringArgBase):
    TYPE_VALUE = "time"


class CMDTimeArg(CMDTimeArgBase, CMDStringArg):
    pass


# uuid
class CMDUuidArgBase(CMDStringArgBase):
    TYPE_VALUE = "uuid"


class CMDUuidArg(CMDUuidArgBase, CMDStringArg):
    pass


# password
class CMDPasswordArgBase(CMDStringArgBase):
    TYPE_VALUE = "password"


class CMDPasswordArg(CMDPasswordArgBase, CMDStringArg):

    prompt = ModelType(CMDPasswordArgPromptInput)


# subscription
class CMDSubscriptionIdArgBase(CMDStringArgBase):
    TYPE_VALUE = "SubscriptionId"


class CMDSubscriptionIdArg(CMDSubscriptionIdArgBase, CMDStringArg):
    pass


# resourceGroupName
class CMDResourceGroupNameArgBase(CMDStringArgBase):
    TYPE_VALUE = "ResourceGroupName"


class CMDResourceGroupNameArg(CMDResourceGroupNameArgBase, CMDStringArg):
    pass


# resourceId
class CMDResourceIdArgBase(CMDStringArgBase):
    TYPE_VALUE = "ResourceId"

    fmt = ModelType(
        CMDResourceIdFormat,
        serialized_name='format',
        deserialize_from='format'
    )

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDResourceIdArgBase)
        arg.fmt = builder.get_fmt()
        return arg


class CMDResourceIdArg(CMDResourceIdArgBase, CMDStringArg):
    pass


# resourceLocation
class CMDResourceLocationArgBase(CMDStringArgBase):
    TYPE_VALUE = "ResourceLocation"

    no_rg_default = CMDBooleanField()  # the default value will not be the location of resource group

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDResourceLocationArgBase)
        arg.no_rg_default = builder.get_resource_location_no_rg_default()
        return arg


class CMDResourceLocationArg(CMDResourceLocationArgBase, CMDStringArg):

    @classmethod
    def build_arg(cls, builder):
        arg = super().build_arg(builder)
        if arg.options == ['location']:
            # add 'l' alias for location arg
            arg.options.append('l')
        return arg


# integer
class CMDIntegerArgBase(CMDArgBase):
    TYPE_VALUE = "integer"

    fmt = ModelType(
        CMDIntegerFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    enum = ModelType(CMDArgEnum)

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDIntegerArgBase)
        arg.fmt = builder.get_fmt()
        arg.enum = builder.get_enum()
        return arg

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.enum:
            self.enum.reformat(**kwargs)
        if self.blank:
            if not isinstance(self.blank.value, int) and not (self.nullable and self.blank.value is None):
                raise exceptions.VerificationError(
                    f"Invalid blank value", details=f"'{self.blank.value}' is not a valid int arg value")


class CMDIntegerArg(CMDIntegerArgBase, CMDArg):

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.default:
            if not isinstance(self.default.value, int) and not (self.nullable and self.default.value is None):
                raise exceptions.VerificationError(
                    f"Invalid default value", details=f"'{self.default.value}' is not a valid int arg value")


# integer32
class CMDInteger32ArgBase(CMDIntegerArgBase):
    TYPE_VALUE = "integer32"


class CMDInteger32Arg(CMDInteger32ArgBase, CMDIntegerArg):
    pass


# integer64
class CMDInteger64ArgBase(CMDIntegerArgBase):
    TYPE_VALUE = "integer64"


class CMDInteger64Arg(CMDInteger64ArgBase, CMDIntegerArg):
    pass


# boolean
class CMDBooleanArgBase(CMDArgBase):
    TYPE_VALUE = "boolean"

    reverse = CMDBooleanField()

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDBooleanArgBase)
        arg.reverse = builder.get_reverse_boolean()
        return arg


class CMDBooleanArg(CMDBooleanArgBase, CMDArg):

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.default:
            if not isinstance(self.default.value, bool) and not (self.nullable and self.default.value is None):
                raise exceptions.VerificationError(
                    f"Invalid default value", details=f"'{self.default.value}' is not a valid boolean arg value")


# float
class CMDFloatArgBase(CMDArgBase):
    TYPE_VALUE = "float"

    fmt = ModelType(
        CMDFloatFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    enum = ModelType(CMDArgEnum)

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDFloatArgBase)
        arg.fmt = builder.get_fmt()
        arg.enum = builder.get_enum()
        return arg

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.enum:
            self.enum.reformat(**kwargs)
        if self.blank:
            if not isinstance(self.blank.value, float) and not (self.nullable and self.blank.value is None):
                if isinstance(self.blank.value, int):
                    # frontend will pass an int value when it doesn't have decimal part
                    self.blank.value = float(self.blank.value)
                else:
                    raise exceptions.VerificationError(
                        f"Invalid blank value", details=f"'{self.blank.value}' is not a valid float arg value")


class CMDFloatArg(CMDFloatArgBase, CMDArg):

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.default:
            if not isinstance(self.default.value, float) and not (self.nullable and self.default.value is None):
                if isinstance(self.default.value, int):
                    # frontend will pass an int value when it doesn't have decimal part
                    self.default.value = float(self.default.value)
                else:
                    raise exceptions.VerificationError(
                        f"Invalid default value", details=f"'{self.default.value}' is not a valid float arg value")


# float32
class CMDFloat32ArgBase(CMDFloatArgBase):
    TYPE_VALUE = "float32"


class CMDFloat32Arg(CMDFloat32ArgBase, CMDFloatArg):
    pass


# float64
class CMDFloat64ArgBase(CMDFloatArgBase):
    TYPE_VALUE = "float64"


class CMDFloat64Arg(CMDFloat64ArgBase, CMDFloatArg):
    pass


class CMDAnyTypeArgBase(CMDArgBase):
    TYPE_VALUE = "any"


class CMDAnyTypeArg(CMDAnyTypeArgBase, CMDArg):
    pass


# object
class CMDObjectArgAdditionalProperties(Model):
    # properties as nodes
    item = CMDArgBaseField()

    # WARNING: this property will be deprecated as CMDAnyTypeArgBase are supported, should directly use CMDAnyTypeArgBase as item value
    any_type = CMDBooleanField(
        serialized_name="anyType",
        deserialize_from="anyType",
    )

    class Options:
        serialize_when_none = False

    @classmethod
    def build_arg_base(cls, builder):
        arg = cls()
        arg.item = builder.get_sub_item()
        arg.any_type = builder.get_any_type()
        return arg

    def reformat(self, **kwargs):
        if self.any_type and not self.item:
            # use CMDAnyTypeArgBase as item value instead
            self.item = CMDAnyTypeArgBase()
            # deprecated the any_type property
            self.any_type = None
        if self.item:
            try:
                self.item.reformat(**kwargs)
            except exceptions.VerificationError as err:
                err.payload['details'] = {
                    "type": "AdditionalProperties",
                    "details": err.payload['details']
                }
                raise err


class CMDObjectArgBase(CMDArgBase):
    TYPE_VALUE = "object"

    fmt = ModelType(
        CMDObjectFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    args = ListType(PolyModelType(CMDArg, allow_subclasses=True))
    additional_props = ModelType(
        CMDObjectArgAdditionalProperties,
        serialized_name="additionalProps",
        deserialize_from="additionalProps",
    )

    # cls definition will not include properties in CMDArg only, such as following properties:
    # var
    # options
    # required
    # stage
    # hide
    # group
    # id_part
    # help
    # default
    cls = CMDClassField()

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDObjectArgBase)
        try:
            arg.fmt = builder.get_fmt()
        except Exception:
            raise
        arg.args = builder.get_sub_args()
        arg.additional_props = builder.get_additional_props()
        if not arg.args and not arg.additional_props:
            # when object arg don't have args or additional_props, set its blank value as empty dict
            arg.blank = CMDArgBlank()
            arg.blank.value = {}
        arg.cls = builder.get_cls()
        return arg

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.args:
            used_options = set()
            for arg in self.args:
                try:
                    arg.reformat(**kwargs)
                except exceptions.VerificationError as err:
                    err.payload['details'] = {
                        "type": "Argument",
                        "options": arg.options,
                        "var": arg.var,
                        "details": err.payload['details']
                    }
                    raise err
                for option in arg.options:
                    if option in used_options:
                        raise exceptions.VerificationError(
                            message=f"Argument option '{option}' duplicated.",
                            details={
                                "type": "Argument",
                                "options": arg.options,
                                "var": arg.var,
                            }
                        )
                    used_options.add(option)
            self.args = sorted(self.args, key=lambda a: a.var)
        if self.additional_props:
            self.additional_props.reformat(**kwargs)

        if self.blank:
            if not isinstance(self.blank.value, dict) and not (self.nullable and self.blank.value is None):
                raise exceptions.VerificationError(
                    f"Invalid blank value", details=f"'{self.blank.value}' is not a valid object arg value")


class CMDObjectArg(CMDObjectArgBase, CMDArg):

    @classmethod
    def build_arg(cls, builder):
        arg = super().build_arg(builder)
        assert isinstance(arg, CMDObjectArg)
        return arg

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.default:
            if not isinstance(self.default.value, dict) and not (self.nullable and self.default.value is None):
                raise exceptions.VerificationError(
                    f"Invalid default value", details=f"'{self.default.value}' is not a valid object arg value")


# array
class CMDArrayArgBase(CMDArgBase):
    TYPE_VALUE = "array"

    fmt = ModelType(
        CMDArrayFormat,
        serialized_name='format',
        deserialize_from='format',
    )

    item = CMDArgBaseField(required=True)

    # cls definition will not include properties in CMDArg only, such as following properties:
    # var
    # options
    # required
    # stage
    # hide
    # group
    # id_part
    # help
    # default
    cls = CMDClassField()

    def _get_type(self):
        return f"{self.TYPE_VALUE}<{self.item.type}>"

    @classmethod
    def build_arg_base(cls, builder):
        arg = super().build_arg_base(builder)
        assert isinstance(arg, CMDArrayArgBase)
        arg.fmt = builder.get_fmt()
        arg.item = builder.get_sub_item()
        arg.cls = builder.get_cls()
        return arg

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.item is None:
            raise exceptions.VerificationError(
                "Invalid array type",
                details="Array item is not defined "
            )
        try:
            self.item.reformat(**kwargs)
        except exceptions.VerificationError as err:
            err.payload['details'] = {
                "type": "ArrayItem",
                "details": err.payload['details']
            }
            raise err

        if self.blank:
            if not isinstance(self.blank.value, list) and not (self.nullable and self.blank.value is None):
                raise exceptions.VerificationError(
                    f"Invalid blank value", details=f"'{self.blank.value}' is not a valid array arg value")


class CMDArrayArg(CMDArrayArgBase, CMDArg):

    singular_options = ListType(
        StringType(),
        serialized_name='singularOptions',
        deserialize_from='singularOptions',
    )  # options to pass element instead of full list

    @classmethod
    def build_arg(cls, builder):
        arg = super().build_arg(builder)
        assert isinstance(arg, CMDArrayArg)
        arg.singular_options = builder.get_singular_options()
        return arg

    def _reformat(self, **kwargs):
        super()._reformat(**kwargs)
        if self.singular_options:
            self.singular_options = sorted(self.singular_options, key=lambda op: (len(op), op))
        if self.default:
            if not isinstance(self.default.value, list) and not (self.nullable and self.default.value is None):
                raise exceptions.VerificationError(
                    f"Invalid default value", details=f"'{self.default.value}' is not a valid array arg value")



import yaml

def _update_config(config, yaml_content):
    for key, value in yaml_content.items():
        if key not in config:
            config[key] = value
            continue
        if isinstance(value, dict):
            _update_config(config[key], value)
        elif isinstance(value, list):
            config[key].extend(value)
        else:
            config[key] = value

def parse_readme_file(readme_path: str):
    """Parse the readme file title and combine basic config in the yaml section."""
    readme_config = {}
    with open(readme_path, 'r', encoding='utf-8') as f:
        content = f.readlines()
    # content.append("```") # append a fake yaml section to make sure the last yaml section is ended
    readme_title = None
    in_yaml_section = False
    yaml_content = []
    for line in content:
        if not readme_title and line.strip().startswith("# ") and not in_yaml_section:
            readme_title = line.strip()[2:].strip()
        if line.strip().startswith("```") and 'yaml' in line:
            condition = line.split('yaml')[1].strip()
            # Do not parse the yaml section if it has the condition
            if not condition:
                in_yaml_section = True
        elif in_yaml_section:
            if line.strip().startswith("```"):
                try:
                    yaml_config = yaml.safe_load("\n".join(yaml_content))
                except Exception as e:
                    raise ValueError(f"Failed to parse autorest config: {e} for readme_file: {readme_path}")
                _update_config(readme_config, yaml_config)
                in_yaml_section = False
                yaml_content = []
            else:
                if line.strip():
                    yaml_content.append(line)
                else:
                    yaml_content.append("")

    return {
        "title": readme_title,
        "config": readme_config
    }




class CloudEnum:
    AzureCloud = 'AzureCloud'
    AzureChinaCloud = 'AzureChinaCloud'
    AzureUSGovernment = 'AzureUSGovernment'
    AzureGermanCloud = 'AzureGermanCloud'

    @classmethod
    def choices(cls):
        return tuple(v for k, v in vars(cls).items() if not k.startswith('_') and isinstance(v, str))



from flask import Blueprint, jsonify, url_for, request
from flask.helpers import send_from_directory
import os

from swagger.controller.specs_manager import SwaggerSpecsManager
from swagger.model.specs import OpenAPIResourceProvider, TypeSpecResourceProvider
from swagger.utils.source import SourceTypeEnum

bp = Blueprint('swagger', __name__, url_prefix='/Swagger/Specs')


@bp.route("/Files/<path:file_path>", methods=("GET",))
def get_file(file_path):
    from utils.config import Config
    if not Config.get_swagger_root():
        return jsonify({"error": "Swagger root not found"}), 404
    return send_from_directory(Config.get_swagger_root(), file_path)


@bp.route("/Stat/<path:path>", methods=("GET",))
def get_path_stat(path):
    from utils.config import Config
    root_dir = Config.get_swagger_root()
    if not root_dir:
        # return 200 status code to avoid print error logs in the client side
        return jsonify({"error": "Swagger root not found"})
    path = os.path.join(root_dir, path)
    if not os.path.exists(path):
        # return 200 status code to avoid print error logs in the client side
        return jsonify({"error": "Path not exist"})
    return jsonify({
        "isDir": os.path.isdir(path),
        "isFile": os.path.isfile(path),
    })


# modules
@bp.route("/<plane>", methods=("GET",))
def get_modules_by(plane):
    specs_manager = SwaggerSpecsManager()
    result = []
    for module in specs_manager.get_modules(plane):
        m = {
            "url": url_for('swagger.get_module', plane=plane, mod_names=module.names),
            "name": module.name,
            "folder": module.folder_path,
        }
        result.append(m)
    return jsonify(result)


@bp.route("/<plane>/<list_path:mod_names>", methods=("GET",))
def get_module(plane, mod_names):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    module = specs_module_manager.module
    result = {
        "url": url_for('swagger.get_module', plane=plane, mod_names=mod_names),
        "name": module.name,
        "folder": module.folder_path,
        "resourceProviders": []
    }
    for rp in specs_module_manager.get_resource_providers():
        if isinstance(rp, OpenAPIResourceProvider):
            result['resourceProviders'].append({
                "url": url_for('swagger.get_openapi_resource_provider', plane=plane, mod_names=mod_names, rp_name=rp.name),
                "name": rp.name,
                "folder": rp.folder_path,
                "type": SourceTypeEnum.OpenAPI,
            })
    return jsonify(result)


# resource providers
@bp.route("/<plane>/<list_path:mod_names>/ResourceProviders", methods=("GET",))
def get_resource_providers_by(plane, mod_names):
    # get query param type in request
    rp_type = request.args.get('type', None)
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    result = []
    for rp in specs_module_manager.get_resource_providers():
        if isinstance(rp, OpenAPIResourceProvider):
            if rp_type and rp_type != 'OpenAPI':
                continue
            result.append({
                "url": url_for('swagger.get_openapi_resource_provider', plane=plane, mod_names=mod_names, rp_name=rp.name),
                "name": rp.name,
                "folder": rp.folder_path,
                "type": SourceTypeEnum.OpenAPI
            })
        elif isinstance(rp, TypeSpecResourceProvider):
            if rp_type and rp_type != 'TypeSpec':
                continue
            result.append({
                "url": url_for('swagger.get_typespec_resource_provider', plane=plane, mod_names=mod_names, rp_name=rp.name),
                "name": rp.name,
                "entryFiles": rp.entry_files,
                "type": SourceTypeEnum.TypeSpec,
            })
    return jsonify(result)


# TODO: may need to add OpenAPI segment in the url
@bp.route("/<plane>/<list_path:mod_names>/ResourceProviders/<rp_name>", methods=("GET",))
def get_openapi_resource_provider(plane, mod_names, rp_name):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    rp = specs_module_manager.get_openapi_resource_provider(rp_name)
    result = {
        "url": url_for('swagger.get_openapi_resource_provider', plane=plane, mod_names=mod_names, rp_name=rp.name),
        "name": rp.name,
        "folder": rp.folder_path,
        "type": SourceTypeEnum.OpenAPI,
        "resources": []
    }
    resource_op_group_map = specs_module_manager.get_grouped_resource_map(rp_name)
    for op_group_name, resource_map in resource_op_group_map.items():
        for resource_id, version_map in resource_map.items():
            rs = {
                "opGroup": op_group_name,
                "url": url_for('swagger.get_resource_in_rp',
                               plane=plane, mod_names=mod_names, rp_name=rp.name, resource_id=resource_id),
                "id": resource_id,
                "versions": []
            }
            for version, resource in version_map.items():
                rs['versions'].append({
                    "url": url_for('swagger.get_resource_version_in_rp',
                                   plane=plane, mod_names=mod_names, rp_name=rp.name,
                                   resource_id=resource.id, version=resource.version),
                    "version": version,
                    "file": resource.file_path,
                    "path": resource.path,
                    "operations": resource.operations
                })
            result['resources'].append(rs)
    return jsonify(result)


@bp.route("/<plane>/<list_path:mod_names>/ResourceProviders/<rp_name>/TypeSpec", methods=("GET",))
def get_typespec_resource_provider(plane, mod_names, rp_name):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    rp = specs_module_manager.get_typespec_resource_provider(rp_name)
    result = {
        "url": url_for('swagger.get_typespec_resource_provider', plane=plane, mod_names=mod_names, rp_name=rp.name),
        "name": rp.name,
        "entryFiles": rp.entry_files,
        "type": "TypeSpec",
    }
    return jsonify(result)


# resources
@bp.route("/<plane>/<list_path:mod_names>/ResourceProviders/<rp_name>/Resources", methods=("GET",))
def get_resources_by(plane, mod_names, rp_name):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    result = []
    rp = specs_module_manager.get_openapi_resource_provider(rp_name)
    resource_op_group_map = specs_module_manager.get_grouped_resource_map(rp_name)
    for op_group_name, resource_map in resource_op_group_map.items():
        for resource_id, version_map in resource_map.items():
            rs = {
                "opGroup": op_group_name,
                "url": url_for('swagger.get_resource_in_rp',
                               plane=plane, mod_names=mod_names, rp_name=rp.name, resource_id=resource_id),
                "id": resource_id,
                "versions": []
            }
            for version, resource in version_map.items():
                rs['versions'].append({
                    "url": url_for('swagger.get_resource_version_in_rp',
                                   plane=plane, mod_names=mod_names, rp_name=rp.name,
                                   resource_id=resource.id, version=resource.version),
                    "id": resource_id,
                    "version": version,
                    "file": resource.file_path,
                    "path": resource.path,
                    "operations": resource.operations
                })
            result.append(rs)
    return jsonify(result)


# resource
@bp.route("/<plane>/<list_path:mod_names>/ResourceProviders/<rp_name>/Resources/<base64:resource_id>",
          methods=("GET",))
def get_resource_in_rp(plane, mod_names, rp_name, resource_id):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    version_map = specs_module_manager.get_resource_version_map(resource_id, rp_name)
    rp = list(version_map.values())[0].resource_provider
    op_group_name = specs_module_manager.get_resource_op_group_name(version_map)
    result = {
        "opGroup": op_group_name,
        "url": url_for('swagger.get_resource_in_rp',
                       plane=plane, mod_names=mod_names, rp_name=rp.name, resource_id=resource_id),
        "id": resource_id,
        "versions": []
    }
    for version, resource in version_map.items():
        result['versions'].append({
            "url": url_for('swagger.get_resource_version_in_rp',
                           plane=plane, mod_names=mod_names, rp_name=rp.name,
                           resource_id=resource.id, version=resource.version),
            "id": resource_id,
            "version": version,
            "file": resource.file_path,
            "path": resource.path,
            "operations": resource.operations
        })
    return jsonify(result)


@bp.route("/<plane>/<list_path:mod_names>/Resources/<base64:resource_id>", methods=("GET",))
def get_resource_in_module(plane, mod_names, resource_id):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    version_map = specs_module_manager.get_resource_version_map(resource_id)
    rp = list(version_map.values())[0].resource_provider
    op_group_name = specs_module_manager.get_resource_op_group_name(version_map)
    result = {
        "opGroup": op_group_name,
        "url": url_for('swagger.get_resource_in_rp',
                       plane=plane, mod_names=mod_names, rp_name=rp.name, resource_id=resource_id),
        "id": resource_id,
        "versions": []
    }
    for version, resource in version_map.items():
        result['versions'].append({
            "url": url_for('swagger.get_resource_version_in_rp',
                           plane=plane, mod_names=mod_names, rp_name=rp.name,
                           resource_id=resource.id, version=resource.version),
            "id": resource_id,
            "version": version,
            "file": resource.file_path,
            "path": resource.path,
            "operations": resource.operations
        })
    return jsonify(result)


# resource version
@bp.route(
    "/<plane>/<list_path:mod_names>/ResourceProviders/<rp_name>/Resources/<base64:resource_id>/V/<base64:version>",
    methods=("GET",)
)
def get_resource_version_in_rp(plane, mod_names, rp_name, resource_id, version):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    resource = specs_module_manager.get_resource_in_version(rp_name, resource_id, version)
    result = {
        "url": url_for('swagger.get_resource_version_in_rp',
                       plane=plane, mod_names=mod_names, rp_name=resource.rp_name,
                       resource_id=resource.id, version=resource.version),
        "id": resource_id,
        "version": version,
        "file": resource.file_path,
        "path": resource.path,
        "operations": resource.operations
    }
    return jsonify(result)


@bp.route("/<plane>/<list_path:mod_names>/Resources/<base64:resource_id>/V/<base64:version>", methods=("GET",))
def get_resource_version_in_module(plane, mod_names, resource_id, version):
    specs_module_manager = SwaggerSpecsManager().get_module_manager(plane, mod_names)
    resource = specs_module_manager.get_resource_in_version(resource_id, version)
    result = {
        "url": url_for('swagger.get_resource_version_in_rp',
                       plane=plane, mod_names=mod_names, rp_name=resource.rp_name,
                       resource_id=resource.id, version=resource.version),
        "id": resource_id,
        "version": version,
        "file": resource.file_path,
        "path": resource.path,
        "operations": resource.operations
    }
    return jsonify(result)



from command.model.configuration import CMDIntegerFormat, CMDStringFormat, CMDFloatFormat, CMDArrayFormat, \
    CMDObjectFormat, CMDSchemaEnum, CMDSchemaEnumItem, CMDDateTimeFormat

from command.model.configuration import CMDSchemaDefault, \
    CMDStringSchema, CMDStringSchemaBase, \
    CMDByteSchema, CMDByteSchemaBase, \
    CMDBinarySchema, CMDBinarySchemaBase, \
    CMDDateSchema, CMDDateSchemaBase, \
    CMDDateTimeSchema, CMDDateTimeSchemaBase, \
    CMDTimeSchema, CMDTimeSchemaBase, \
    CMDPasswordSchema, CMDPasswordSchemaBase, \
    CMDDurationSchema, CMDDurationSchemaBase, \
    CMDUuidSchema, CMDUuidSchemaBase, \
    CMDResourceIdSchema, CMDResourceIdSchemaBase, \
    CMDIntegerSchema, CMDIntegerSchemaBase, \
    CMDInteger32Schema, CMDInteger32SchemaBase, \
    CMDInteger64Schema, CMDInteger64SchemaBase, \
    CMDBooleanSchema, CMDBooleanSchemaBase, \
    CMDFloatSchema, CMDFloatSchemaBase, \
    CMDFloat32Schema, CMDFloat32SchemaBase, \
    CMDFloat64Schema, CMDFloat64SchemaBase, \
    CMDObjectSchema, CMDObjectSchemaBase, \
    CMDArraySchema, CMDArraySchemaBase, \
    CMDClsSchema, CMDClsSchemaBase, \
    CMDAnyTypeSchema, CMDAnyTypeSchemaBase, \
    CMDHttpResponseJsonBody
from command.controller.workspace_cfg_editor import WorkspaceCfgEditor

from swagger.model.specs._utils import operation_id_separate
from swagger.utils import exceptions
from .fields import MutabilityEnum
from .response import Response
from .schema import ReferenceSchema
from .x_ms_pageable import XmsPageable
from functools import reduce
from utils.case import to_camel_case
import logging
import re

logger = logging.getLogger("aaz")


class CMDBuilder:

    def __init__(self, path, method=None, mutability=None, in_base=False, frozen=False, parent_ids=None, cls_definitions=None, parameterized_host=None):
        self.path = path
        self.method = method
        self.mutability = mutability
        self.in_base = in_base
        self.frozen = frozen
        self.read_only = False
        self.id = None    # used to find loop
        self.parent_ids = parent_ids or []
        self.cls_definitions = {} if cls_definitions is None else cls_definitions
        self.parameterized_host = parameterized_host

    def __call__(self, schema, **kwargs):
        sub_builder = CMDBuilder(
            path=kwargs.pop('path', self.path),
            method=kwargs.pop('method', self.method),
            mutability=kwargs.pop('mutability', self.mutability),
            in_base=kwargs.pop('in_base', self.in_base),
            frozen=kwargs.pop('frozen', self.frozen),
            parent_ids=[*self.parent_ids, self.id],
            cls_definitions=kwargs.pop('cls_definitions', self.cls_definitions),
            parameterized_host=kwargs.pop('parameterized_host', self.parameterized_host)
        )
        if getattr(schema, 'read_only', None):
            sub_builder.read_only = True
        if not sub_builder.frozen:
            if sub_builder.read_only:
                if self.mutability != MutabilityEnum.Read:
                    sub_builder.frozen = True
            elif getattr(schema, 'x_ms_mutability', None):
                if self.mutability not in schema.x_ms_mutability:
                    sub_builder.frozen = True
        if hasattr(schema, 'traces'):
            sub_builder.id = (schema.traces, sub_builder.mutability, sub_builder.frozen)
            if sub_builder.id in sub_builder.parent_ids:
                if len(schema.traces) == 3:
                    # make sure the trace is reference definition, the trace should be [file_path, 'definitions', name]
                    raise exceptions.InvalidSwaggerValueError(
                            msg="Find invalid reference loop",
                            key=sub_builder.id,
                            value=sub_builder.parent_ids.index(sub_builder.id),
                        )
        return schema.to_cmd(sub_builder, **kwargs)

    def find_traces(self, traces):
        assert traces is not None
        for parent_id in self.parent_ids:
            if parent_id is None:
                continue
            parent_traces, mutability, frozen = parent_id
            if parent_traces == traces:
                return True
        return False

    def build_schema(self, schema):
        schema_type = getattr(schema, 'type', None)
        if schema_type == "string":
            if schema.format is None or schema.format == "uri":
                if self.in_base:
                    model = CMDStringSchemaBase()
                else:
                    model = CMDStringSchema()
            elif schema.format == "byte":
                if self.in_base:
                    model = CMDByteSchemaBase()
                else:
                    model = CMDByteSchema()
            elif schema.format == "binary":
                if self.in_base:
                    model = CMDBinarySchemaBase()
                else:
                    model = CMDBinarySchema()
            elif schema.format == "date":
                if self.in_base:
                    model = CMDDateSchemaBase()
                else:
                    model = CMDDateSchema()
            elif schema.format == "date-time" or schema.format == "date-time-rfc1123":
                if self.in_base:
                    model = CMDDateTimeSchemaBase()
                else:
                    model = CMDDateTimeSchema()
            elif schema.format == "time":
                if self.in_base:
                    model = CMDTimeSchemaBase()
                else:
                    model = CMDTimeSchema()
            elif schema.format == "password":
                if self.in_base:
                    model = CMDPasswordSchemaBase()
                else:
                    model = CMDPasswordSchema()
            elif schema.format == "duration":
                if self.in_base:
                    model = CMDDurationSchemaBase()
                else:
                    model = CMDDurationSchema()
            elif schema.format == "uuid":
                if self.in_base:
                    model = CMDUuidSchemaBase()
                else:
                    model = CMDUuidSchema()
            elif schema.format == "arm-id":
                if self.in_base:
                    model = CMDResourceIdSchemaBase()
                else:
                    model = CMDResourceIdSchema()
            else:
                raise exceptions.InvalidSwaggerValueError(
                    f"format is not supported", key=getattr(schema, "traces", None), value=[schema_type, schema.format])
        elif schema_type == "integer":
            if schema.format is None:
                if self.in_base:
                    model = CMDIntegerSchemaBase()
                else:
                    model = CMDIntegerSchema()
            elif schema.format == "int32":
                if self.in_base:
                    model = CMDInteger32SchemaBase()
                else:
                    model = CMDInteger32Schema()
            elif schema.format == "int64":
                if self.in_base:
                    model = CMDInteger64SchemaBase()
                else:
                    model = CMDInteger64Schema()
            else:
                raise exceptions.InvalidSwaggerValueError(
                    f"format is not supported", key=getattr(schema, "traces", None), value=[schema_type, schema.format])
        elif schema_type == "boolean":
            if schema.format is None:
                if self.in_base:
                    model = CMDBooleanSchemaBase()
                else:
                    model = CMDBooleanSchema()
            else:
                raise exceptions.InvalidSwaggerValueError(
                    f"format is not supported", key=getattr(schema, "traces", None), value=[schema_type, schema.format])
        elif schema_type == "number":
            if schema.format is None:
                if self.in_base:
                    model = CMDFloatSchemaBase()
                else:
                    model = CMDFloatSchema()
            elif schema.format == "float":
                if self.in_base:
                    model = CMDFloat32SchemaBase()
                else:
                    model = CMDFloat32Schema()
            elif schema.format == "double":
                if self.in_base:
                    model = CMDFloat64SchemaBase()
                else:
                    model = CMDFloat64Schema()
            else:
                raise exceptions.InvalidSwaggerValueError(
                    f"format is not supported", key=getattr(schema, "traces", None), value=[schema_type, schema.format])
        elif schema_type == "array":
            if schema.format is None:
                if self.in_base:
                    model = CMDArraySchemaBase()
                else:
                    model = CMDArraySchema()
            else:
                raise exceptions.InvalidSwaggerValueError(
                    f"format is not supported", key=getattr(schema, "traces", None), value=[schema_type, schema.format])
        elif schema_type == "object" or getattr(schema, "properties", None) or \
                getattr(schema, "additional_properties", None):
            if schema.format is None:
                if self.in_base:
                    model = CMDObjectSchemaBase()
                else:
                    model = CMDObjectSchema()
            else:
                raise exceptions.InvalidSwaggerValueError(
                    f"format is not supported", key=getattr(schema, "traces", None), value=[schema_type, schema.format])
        # for swagger schema only
        elif getattr(schema, "all_of", None) is not None:
            model = self.build_schema(schema.all_of[0])
        elif getattr(schema, "ref_instance", None) is not None:
            model = self.build_schema(schema.ref_instance)
        elif schema_type is None:  # according to the swagger spec, the type is any if not defined
            if self.in_base:
                model = CMDAnyTypeSchemaBase()
            else:
                model = CMDAnyTypeSchema()
        else:
            raise exceptions.InvalidSwaggerValueError(
                f"type is not supported", key=getattr(schema, "traces", None), value=[schema_type])

        model.read_only = self.read_only
        model.frozen = self.frozen
        return model

    def _get_cls_definition_name(self, schema):
        assert isinstance(schema, ReferenceSchema)
        schema_cls_name = f"{to_camel_case(schema.ref.split('/')[-1].replace('.', ' '))}_{self.mutability}"
        if self.mutability != MutabilityEnum.Read:
            if self.read_only:
                schema_cls_name += "_read"
        if self.frozen:
            schema_cls_name += "_frozen"
        return schema_cls_name

    def register_cls_definition(self, schema, support_cls_schema, **kwargs):
        name = self._get_cls_definition_name(schema)
        if self.frozen:
            if support_cls_schema:
                if self.in_base:
                    model = CMDClsSchemaBase()
                else:
                    model = CMDClsSchema()
                model.read_only = self.read_only
                model.frozen = self.frozen
                model._type = f"@{name}"
            else:
                model = self(schema.ref_instance, **kwargs)
                model.read_only = self.read_only
            return model

        if name not in self.cls_definitions:
            if support_cls_schema:
                # register in cls_definitions first in case of loop reference below
                self.cls_definitions[name] = {"count": 1}
                model = self(schema.ref_instance, **kwargs)
                model.read_only = self.read_only
                if isinstance(model, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                    # Important: only support object and array schema to defined as cls
                    # when self.cls_definitions[name]['count'] > 1, the loop reference exist
                    self.cls_definitions[name]['model'] = model
                else:
                    del self.cls_definitions[name]
            else:
                model = self(schema.ref_instance, **kwargs)
                model.read_only = self.read_only
        else:
            if support_cls_schema:
                self.cls_definitions[name]['count'] += 1
                if self.in_base:
                    model = CMDClsSchemaBase()
                else:
                    model = CMDClsSchema()
                model.read_only = self.read_only
                if 'model' in self.cls_definitions[name]:
                    # need to combine with the model frozen, especially for _create model with all ready_only properties
                    model.frozen = self.frozen or self.cls_definitions[name]['model'].frozen
                    # link implement
                    model.implement = self.cls_definitions[name]['model']
                else:
                    # valid reference loop
                    model.frozen = self.frozen
                model._type = f"@{name}"
            else:
                if 'model' not in self.cls_definitions[name]:
                    raise exceptions.InvalidSwaggerValueError(
                        msg="Find invalid reference loop",
                        key=schema.traces,
                        value=name
                    )
                model = self(schema.ref_instance, **kwargs)
                model.read_only = self.read_only
        return model

    def get_cls_definition_model(self, model):
        assert isinstance(model, CMDClsSchemaBase)
        name = model.type[1:]
        return self.cls_definitions[name]['model']

    def parse_parameterized_host_path(self):
        """ parse the path of the host template and parameters used in it. """
        if not self.parameterized_host or not self.parameterized_host.host_template:
            return None, None
        result = re.fullmatch(r"^(.*://)?[^/]*(/.*)$", self.parameterized_host.host_template)
        host_path = result[2] if result else None
        if not host_path or host_path == '/':
            return None, None
        parameter_names = set()
        for r in re.finditer(r"\{([^{}]*)}", host_path):
            parameter_names.add(r[1])
        parameters = []
        if self.parameterized_host.parameters:
            for param in self.parameterized_host.parameters:
                if param.name in parameter_names:
                    parameters.append(param)
        return host_path, parameters

    @staticmethod
    def setup_enum(model, schema):
        if not schema.enum and not (getattr(schema, 'x_ms_enum', None) and schema.x_ms_enum.values):
            return
        enum = CMDSchemaEnum()
        enum.items = []
        if schema.x_ms_enum and schema.x_ms_enum.values:
            for v in schema.x_ms_enum.values:
                item = CMDSchemaEnumItem()
                item.value = v.value
                if v.name:
                    # TODO: the name should be used as display name for argument
                    pass
                enum.items.append(item)
        elif schema.enum:
            for v in schema.enum:
                item = CMDSchemaEnumItem()
                item.value = v
                enum.items.append(item)

        model.enum = enum

    def setup_fmt(self, model, schema):
        if not hasattr(model, 'fmt'):
            return
        fmt = None
        if isinstance(model, CMDDateTimeSchemaBase):
            # CMDDateTimeSchemaBase is subclass of CMDStringSchemaBase, it needs to be added before following string fmt
            fmt = self.build_cmd_datetime_format(schema)
        elif isinstance(model, CMDStringSchemaBase):
            fmt = self.build_cmd_string_format(schema)
        elif isinstance(model, CMDIntegerSchemaBase):
            fmt = self.build_cmd_integer_format(schema)
        elif isinstance(model, CMDBooleanSchemaBase):
            # TODO:
            pass
        elif isinstance(model, CMDFloatSchemaBase):
            fmt = self.build_cmd_float_format(schema)
        elif isinstance(model, CMDArraySchemaBase):
            fmt = self.build_cmd_array_format(schema)
        elif isinstance(model, CMDObjectSchemaBase):
            fmt = self.build_cmd_object_format(schema)

        model.fmt = fmt or model.fmt

    @staticmethod
    def setup_default(model, schema):
        if getattr(schema, 'x_ms_client_default', None) is not None:
            model.default = CMDSchemaDefault()
            model.default.value = schema.x_ms_client_default
        elif schema.default is not None:
            model.default = CMDSchemaDefault()
            model.default.value = schema.default

    @staticmethod
    def setup_nullable(model, schema):
        if getattr(schema, 'x_nullable', False):
            model.nullable = True

    @staticmethod
    def setup_description(model, schema):
        if schema.description:
            model.description = schema.description
        elif getattr(schema, 'title', None):
            model.description = schema.title
        elif getattr(schema, 'summary', None):
            model.description = schema.summary

    @staticmethod
    def setup_secret(model, schema):
        if getattr(schema, 'x_ms_secret', False):
            model.secret = schema.x_ms_secret

    @staticmethod
    def build_cmd_string_format(schema):
        fmt_assigned = False
        fmt = CMDStringFormat()

        if schema.pattern is not None:
            try:
                _ = re.compile(schema.pattern)  # verify schema pattern
                fmt.pattern = schema.pattern
                fmt_assigned = True
            except Exception as err:
                logger.warning('Invalid regex expression: traces: {}, pattern: {}'.format(str([schema.traces]), schema.pattern))
        if schema.max_length is not None:
            fmt.max_length = schema.max_length
            fmt_assigned = True
        if schema.min_length is not None:
            fmt.min_length = schema.min_length
            fmt_assigned = True

        if not fmt_assigned:
            return None
        return fmt

    @staticmethod
    def build_cmd_datetime_format(schema):
        fmt = CMDDateTimeFormat()
        if schema.format == "date-time-rfc1123":
            fmt.protocol = "rfc"
        return fmt

    @staticmethod
    def build_cmd_integer_format(schema):
        fmt_assigned = False
        fmt = CMDIntegerFormat()

        if schema.maximum is not None:
            fmt.maximum = int(schema.maximum)
            if schema.exclusive_maximum and fmt.maximum == schema.maximum:
                fmt.maximum -= 1
            fmt_assigned = True

        if schema.minimum is not None:
            fmt.minimum = int(schema.minimum)
            if schema.exclusive_minimum and fmt.minimum == schema.minimum:
                fmt.minimum += 1
            fmt_assigned = True

        if schema.multiple_of is not None:
            fmt.multiple_of = schema.multiple_of
            fmt_assigned = True

        if not fmt_assigned:
            return None
        return fmt

    @staticmethod
    def build_cmd_float_format(schema):
        fmt_assigned = False
        fmt = CMDFloatFormat()

        if schema.maximum is not None:
            fmt.maximum = schema.maximum
            if schema.exclusive_maximum:
                fmt.exclusive_maximum = True
            fmt_assigned = True

        if schema.minimum is not None:
            fmt.minimum = int(schema.minimum)
            if schema.exclusive_minimum:
                fmt.exclusive_minimum = True
            fmt_assigned = True

        if schema.multiple_of is not None:
            fmt.multiple_of = schema.multiple_of
            fmt_assigned = True

        if not fmt_assigned:
            return None
        return fmt

    @staticmethod
    def build_cmd_array_format(schema):
        fmt_assigned = False
        fmt = CMDArrayFormat()

        if schema.unique_items:
            fmt.unique = True
            fmt_assigned = True

        if schema.max_items is not None:
            fmt.max_length = schema.max_items
            fmt_assigned = True

        if schema.min_items is not None:
            fmt.min_length = schema.min_items
            fmt_assigned = True

        if getattr(schema, "collection_format", None) is not None and schema.collection_format != 'csv':
            fmt.str_format = schema.collection_format
            fmt_assigned = True

        if not fmt_assigned:
            return None
        return fmt

    @staticmethod
    def build_cmd_object_format(schema):
        fmt_assigned = False
        fmt = CMDObjectFormat()

        if schema.max_properties is not None:
            fmt.max_properties = schema.max_properties
            fmt_assigned = True
        if schema.min_properties is not None:
            fmt.min_properties = schema.min_properties
            fmt_assigned = True

        if not fmt_assigned:
            return None
        return fmt

    @staticmethod
    def classify_responses(schema):
        success_responses = []
        success_202_response = None
        success_204_response = None
        redirect_responses = []
        error_responses = []

        if 'default' in schema.responses:
            resp = schema.responses['default']
            error_responses.append((None, resp))

        for code, resp in schema.responses.items():
            if code == "default":
                continue
            status_code = int(code)
            if status_code < 300:
                if status_code == 202:
                    success_202_response = ({status_code}, resp)
                elif status_code == 204:
                    success_204_response = ({status_code}, resp)
                else:
                    # success
                    find_match = False
                    for status_codes, p_resp in success_responses:
                        if p_resp.schema == resp.schema:
                            status_codes.add(status_code)
                            find_match = True
                            break
                    if not find_match:
                        success_responses.append(({status_code}, resp))
            elif status_code < 400:
                # redirect
                find_match = False
                for status_codes, p_resp in redirect_responses:
                    if p_resp.schema == resp.schema:
                        status_codes.add(status_code)
                        find_match = True
                        break
                if not find_match:
                    redirect_responses.append(({status_code}, resp))
            else:
                # error
                find_match = False
                for status_codes, p_resp in error_responses:
                    if p_resp.schema == resp.schema:
                        if status_codes is not None:  # ignore to add into default response
                            status_codes.add(status_code)
                        find_match = True
                if not find_match:
                    error_responses.append(({status_code}, resp))

        if len(success_responses) >= 2:
            raise exceptions.InvalidSwaggerValueError(
                msg="Multi Schema for success responses",
                key=[schema.traces],
                value=[status_codes for status_codes, _ in success_responses]
            )
        if len(redirect_responses) >= 2:
            raise exceptions.InvalidSwaggerValueError(
                msg="Multi Schema for redirect responses",
                key=[schema.traces],
                value=[status_codes for status_codes, _ in redirect_responses]
            )
        if len(error_responses) >= 3:
            raise exceptions.InvalidSwaggerValueError(
                msg="Multi Schema for error responses",
                key=[schema.traces],
                value=[status_codes for status_codes, _ in error_responses]
            )

        if success_202_response is not None:
            # append 202 Long Running response at the end of success response
            success_responses.append(success_202_response)
        if success_204_response is not None:
            # append 204 No Content response at the end of success response
            success_responses.append(success_204_response)

        success_codes = reduce(lambda x, y: x | y, [codes for codes, _ in success_responses], set())
        if schema.x_ms_long_running_operation and not success_codes & {200, 201}:
            if lro_schema := schema.x_ms_lro_final_state_schema:
                lro_response = Response()
                lro_response.description = "Response schema for long-running operation."
                lro_response.schema = lro_schema

                success_responses.append(({200, 201}, lro_response))  # use `final-state-schema` as response

            elif operation_id_separate(schema.operation_id)[-1][0] == "delete":
                lro_response = Response()
                lro_response.description = "Response schema for long-running operation."
                success_responses.append(({200, 201}, lro_response))
            else:
                logger.warning(f"No response schema for long-running-operation: {schema.operation_id}.")

        # # default response
        # if 'default' not in error_responses and len(error_responses) == 1:
        #     p_resp, p_model = [*error_responses.values()][0]
        #     if p_model.body is not None:
        #         # use the current error response as default
        #         p_model.status_codes = None
        #         error_responses = {
        #             "default": (p_resp, p_model)
        #         }
        # if 'default' not in error_responses:
        #     raise exceptions.InvalidSwaggerValueError(
        #         msg="Miss default response",
        #         key=self.traces,
        #         value=[path, method]
        #     )

        return success_responses, redirect_responses, error_responses

    def apply_cls_definitions(self, *cmd_ops):
        for name, definition in self.cls_definitions.items():
            if definition['count'] > 1:
                definition['model'].cls = name
        check = True
        while check:
            check = False
            schema_cls_register_map = {}
            for cmd_op in cmd_ops:
                cmd_op.register_cls(cls_register_map=schema_cls_register_map)
            
            for name, cls_register in schema_cls_register_map.items():
                if cls_register.get('implement', None):
                    continue

                new_schema = None

                for parent, schema, _ in WorkspaceCfgEditor.iter_schema_cls_reference_in_operations(cmd_ops, name):
                    if schema.frozen:
                        continue

                    if new_schema is not None:
                        schema.implement = new_schema
                        new_schema.cls = name
                        continue

                    # the linked implement will be lost when it's rebuild in the get_unwrapped()
                    schema.implement = self.cls_definitions[name]['model']
                    new_schema = schema.get_unwrapped()
                    assert new_schema is not None
                    new_schema.cls = None
                    WorkspaceCfgEditor.replace_schema(parent, schema, new_schema)
                    self.cls_definitions[name]['model'] = new_schema
                    # because the schema is updated, need to re-check the loop
                    check = True

    def get_pageable(self, path_item, op):
        pageable = getattr(path_item, self.method).x_ms_pageable
        if pageable is None and self.method == "get":
            # some list operation may miss pageable
            for resp in op.http.responses:
                if resp.is_error:
                    continue
                if not isinstance(resp.body, CMDHttpResponseJsonBody):
                    continue
                if not isinstance(resp.body.json.schema, CMDObjectSchemaBase):
                    continue
                body_schema = resp.body.json.schema
                if not body_schema.props:
                    continue

                has_value = False
                has_next_link = False
                for prop in body_schema.props:
                    if prop.name == "value" and isinstance(prop, CMDArraySchema):
                        has_value = True
                    if prop.name == "nextLink" and isinstance(prop, CMDStringSchema):
                        has_next_link = True
                if has_value and has_next_link:
                    pageable = XmsPageable()
                    pageable.next_link_name = "nextLink"

        return pageable



from command.model.configuration import CMDStageField, CMDHelp, CMDCommandExample
from command.model.configuration._fields import CMDCommandNameField, CMDVersionField
from schematics.models import Model
from schematics.common import NOT_NONE
from schematics.types import ModelType, ListType, DictType

from ._resource import CMDSpecsResource


class CMDSpecsSimpleCommand(Model):
    names = ListType(field=CMDCommandNameField(), min_size=1, required=True)  # full name of a command

    class Options:
        serialize_when_none = False


class CMDSpecsSimpleCommandGroup(Model):
    names = ListType(field=CMDCommandNameField(), min_size=1, required=True)  # full name of a command group
    command_groups = DictType(
        field=ModelType("CMDSpecsSimpleCommandGroup"),
        serialized_name="commandGroups",
        deserialize_from="commandGroups",
        export_level=NOT_NONE,
    )
    commands = DictType(
        field=ModelType(CMDSpecsSimpleCommand),
        export_level=NOT_NONE,
    )

    class Options:
        serialize_when_none = False


class CMDSpecsSimpleCommandTree(Model):
    root = ModelType(
        CMDSpecsSimpleCommandGroup
    )  # the root node

    class Options:
        serialize_when_none = False


class CMDSpecsCommandVersion(Model):
    name = CMDVersionField(required=True)
    stage = CMDStageField()
    resources = ListType(ModelType(CMDSpecsResource), required=True, min_size=1)
    examples = ListType(ModelType(CMDCommandExample))

    class Options:
        serialize_when_none = False


class CMDSpecsCommand(Model):
    names = ListType(field=CMDCommandNameField(), min_size=1, required=True)  # full name of a command
    help = ModelType(CMDHelp, required=True)
    versions = ListType(ModelType(CMDSpecsCommandVersion), required=True, min_size=1)

    class Options:
        serialize_when_none = False


class CMDSpecsCommandGroup(Model):
    names = ListType(field=CMDCommandNameField(), min_size=1, required=True)  # full name of a command group
    help = ModelType(CMDHelp)

    command_groups = DictType(
        field=ModelType("CMDSpecsCommandGroup"),
        serialized_name="commandGroups",
        deserialize_from="commandGroups"
    )
    commands = DictType(
        field=ModelType(CMDSpecsCommand)
    )

    class Options:
        serialize_when_none = False


class CMDSpecsCommandTree(Model):
    root = ModelType(
        CMDSpecsCommandGroup
    )  # the root node

    class Options:
        serialize_when_none = False



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from cli.model.common._fields import CLICommandNameField
from schematics.models import Model
from schematics.types import ModelType, ListType, DictType

from ._command import CLIViewCommand


class CLIViewCommandGroup(Model):
    names = ListType(field=CLICommandNameField(), min_size=1, required=True)  # full name of a command group

    command_groups = DictType(
        field=ModelType("CLIViewCommandGroup"),
        serialized_name="commandGroups",
        deserialize_from="commandGroups"
    )
    commands = DictType(
        field=ModelType(CLIViewCommand)
    )
    wait_command = ModelType(
        CLIViewCommand,
        serialized_name="waitCommand",
        deserialize_from="waitCommand"
    )

    class Options:
        serialize_when_none = False



import os
import logging
import re

logger = logging.getLogger('aaz')


def map_path_2_repo(path):
    path = os.path.normpath(path)
    ems = path.split(os.sep)
    idx = ems.index('specification')
    ems = ems[idx:]
    return '/'.join(['https://github.com/Azure/azure-rest-api-specs/tree/master', *ems])


def camel_case_to_snake_case(name, separator='_'):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1' + separator + r'\2', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1' + separator + r'\2', name).lower()
    return name


def operation_id_separate(op_id):
    value = op_id.strip()
    value = value.replace('-', '_')
    value = value.replace('.', '_')
    value = value.replace(' ', '')
    parts = value.split('_')
    for idx in range(len(parts)):
        part = parts[idx]
        part = camel_case_to_snake_case(part, separator='_')
        parts[idx] = [p for p in part.split('_') if p]
    return parts


def get_url_path_valid_parts(url_path, rp_name):
    """valid parts of url path is a section of url without subscription, resource group and providers"""
    url_path = url_path.split('?')[0]
    parts = url_path.split("/")

    provider_idx = None
    for idx, part in enumerate(parts):
        if part.lower() == rp_name.lower():
            provider_idx = idx

    if provider_idx is None:
        for idx, part in enumerate(parts):
            if part.lower() == "providers":
                provider_idx = idx + 1

    if provider_idx is not None:
        valid_idx = provider_idx
    else:
        valid_idx = 0
        for idx, part in enumerate(parts):
            if part.lower() == "subscriptions":
                valid_idx = idx + 1
            if part.lower() == "resourcegroups":
                valid_idx = idx + 1

    parts = [part for part in parts[valid_idx:] if part not in ('locations', )]
    return parts



import logging
import os
import re

from command.model.configuration import CMDHelp, CMDCommandExample
from command.model.specs import CMDSpecsCommandGroup, CMDSpecsCommand, CMDSpecsResource, CMDSpecsCommandVersion, \
    CMDSpecsCommandTree
from command.model.specs._command_tree import CMDSpecsSimpleCommand, CMDSpecsSimpleCommandGroup, \
    CMDSpecsSimpleCommandTree
from utils import exceptions

logger = logging.getLogger(__name__)


def _build_simple_command(names):
    # uri = '/Commands/' + '/'.join(names[:-1]) + f'/_{names[-1]}.md'
    command = CMDSpecsSimpleCommand()
    command.names = names
    return command


def _build_simple_command_group(names, aaz_path):
    """
    Build Simple Command Group from directory
    """
    rel_names = names
    if len(names) == 1 and names[0] == 'aaz':
        rel_names = []
    # uri = '/Commands/' + '/'.join(rel_names) + f'/readme.md'
    full_path = os.path.join(aaz_path, 'Commands', *rel_names)
    if rel_names and not os.path.exists(os.path.join(full_path, 'readme.md')):
        return None
    commands = {}
    command_groups = {}
    for dir in os.listdir(full_path):
        if os.path.isfile(os.path.join(full_path, dir)):
            if not dir.endswith('.md') or dir == 'readme.md':
                continue
            command_name = dir[1:-3]
            commands[command_name] = _build_simple_command(rel_names + [command_name])
        else:
            cg_name = dir
            group = _build_simple_command_group(rel_names + [cg_name], aaz_path)
            if group:
                command_groups[cg_name] = group
    cg = CMDSpecsSimpleCommandGroup()
    cg.names = names
    cg.commands = commands
    cg.command_groups = command_groups
    return cg


def build_simple_command_tree(aaz_path):
    root = _build_simple_command_group(['aaz'], aaz_path)
    tree = CMDSpecsSimpleCommandTree()
    tree.root = root
    tree.validate()
    return tree


class CMDSpecsPartialCommandGroup:
    def __init__(self, names, short_help, uri, aaz_path):
        self.names = names
        self.short_help = short_help
        self.uri = uri
        self.aaz_path = aaz_path

    @classmethod
    def parse_command_group_info(cls, info, cg_names, aaz_path):
        lines = info.splitlines(keepends=False)

        cg = CMDSpecsCommandGroup()
        _, _, remaining_lines = cls._parse_title(lines)
        cg.names = list(cg_names) or ["aaz"]
        cg.help, remaining_lines = cls._parse_help(remaining_lines)
        cg.command_groups, remaining_lines = cls._parse_groups(remaining_lines, cg_names, aaz_path)
        cg.commands, _ = cls._parse_commands(remaining_lines, cg_names, aaz_path)

        return cg

    @classmethod
    def _parse_title(cls, lines):
        assert len(lines) > 0 and lines[0].startswith('# ')
        title_line = lines[0].strip()
        if not title_line.endswith('_'):    # root
            return None, None, lines[2:]
        _, category, names = title_line.split(maxsplit=2)
        return category[1: -1], names[1: -1], lines[2:]

    @classmethod
    def _parse_help(cls, lines):
        assert len(lines) > 0
        if lines[0].startswith('## '):  # root
            return None, lines
        short_help, remaining_lines = cls._read_until(lines, lambda line: not line)
        short_help = '\n'.join(short_help)
        remaining_lines = cls._del_empty(remaining_lines)
        long_help, remaining_lines = cls._read_until(remaining_lines, lambda line: line.startswith('## '))
        long_help = long_help[:-1] if long_help and not long_help[-1] else long_help    # Delete last line if empty
        return CMDHelp(raw_data={'short': short_help, 'lines': long_help or None}), remaining_lines

    @classmethod
    def _parse_groups(cls, lines: list[str], cg_names, aaz_path):
        if lines and lines[0] in ['## Groups', '## Subgroups']:
            groups = []
            remaining_lines = lines[2:]
            while remaining_lines and not remaining_lines[0].startswith('## '):
                group, remaining_lines = cls._parse_item(
                    remaining_lines, CMDSpecsPartialCommandGroup, cg_names, aaz_path)
                groups.append((group.names[-1], group))
            return CMDSpecsCommandGroupDict(groups), remaining_lines
        return CMDSpecsCommandGroupDict([]), lines

    @classmethod
    def _parse_commands(cls, lines: list[str], cg_names, aaz_path):
        if lines and lines[0] in ['## Commands']:
            commands = []
            remaining_lines = lines[2:]
            while remaining_lines and not remaining_lines[0].startswith('## '):
                command, remaining_lines = cls._parse_item(
                    remaining_lines, CMDSpecsPartialCommand, cg_names, aaz_path)
                commands.append((command.names[-1], command))
            return CMDSpecsCommandDict(commands), remaining_lines
        return CMDSpecsCommandDict([]), lines

    @classmethod
    def _parse_item(cls, lines, item_cls, cg_names, aaz_path):
        assert len(lines) > 1
        name_line = lines[0]
        assert name_line.startswith('- [')
        name = name_line[3:].split(']')[0]
        uri = name_line.split('(')[-1].split(')')[0]
        short_help, remaining_lines = cls._read_until(lines[1:], lambda line: not line)
        remaining_lines = cls._del_empty(remaining_lines)
        short_help = '\n'.join(short_help)
        return item_cls(names=[*cg_names, name], short_help=short_help, uri=uri, aaz_path=aaz_path), remaining_lines

    @classmethod
    def _read_until(cls, lines, predicate):
        result = []
        for idx in range(0, len(lines)):
            if predicate(lines[idx]):
                return result, lines[idx:]
            result.append(lines[idx])
        return result, []

    @classmethod
    def _del_empty(cls, lines):
        for idx in range(0, len(lines)):
            if lines[idx]:
                return lines[idx:]
        return lines

    def load(self):
        with open(self.aaz_path + self.uri, "r", encoding="utf-8") as f:
            content = f.read()
            if self.names and self.names[0] == "aaz":
                names = self.names[1:]
            else:
                names = self.names
            cg = self.parse_command_group_info(content, names, self.aaz_path)
            return cg


class CMDSpecsCommandDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, __key):
        command = super().__getitem__(__key)
        if isinstance(command, CMDSpecsPartialCommand):
            command = command.load()
            if command:
                super().__setitem__(__key, command)
        return command

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def values(self):
        for key in self.keys():
            yield self[key]

    def get_raw_item(self, key):
        return super().get(key)

    def raw_values(self):
        return super().values()

    def raw_items(self):
        return super().items()


class CMDSpecsCommandGroupDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, __key):
        cg = super().__getitem__(__key)
        if isinstance(cg, CMDSpecsPartialCommandGroup):
            cg = cg.load()
            if cg:
                super().__setitem__(__key, cg)
        return cg

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def values(self):
        for key in self.keys():
            yield self[key]

    def get_raw_item(self, key):
        return super().get(key)

    def raw_values(self):
        return super().values()

    def raw_items(self):
        return super().items()


class CMDSpecsPartialCommand:
    _COMMAND_INFO_RE = r"# \[Command\] _(?P<group_name>[A-Za-z0-9- ]+)_\n" \
                       r"\n(?P<short_help>(.+\n)+)\n?" \
                       r"(?P<lines_help>\n(^[^#].*\n)+)?" \
                       r"\n## Versions\n" \
                       r"(?P<versions>(\n### \[(?P<version_name>[a-zA-Z0-9-]+)\]\(.*\) \*\*.*\*\*\n" \
                       r"\n(<!-- .* -->\n)+" \
                       r"(?P<examples>\n#### examples\n" \
                       r"(\n- (?P<example_desc>.*)\n" \
                       r"    ```bash\n" \
                       r"(        (?P<example_cmd>.*)\n)+" \
                       r"    ```\n)+)?)+)"
    COMMAND_INFO_RE = re.compile(_COMMAND_INFO_RE, re.MULTILINE)
    _RESOURCE_INFO_RE = r"<!-- (?P<plane>\S+) (?P<id>\S+) (?P<version>\S+) ((?P<subresource>\S+) )?-->\n"
    RESOURCE_INFO_RE = re.compile(_RESOURCE_INFO_RE, re.MULTILINE)
    _VERSION_INFO_RE = r"### \[(?P<version_name>[a-zA-Z0-9-]+)\]\(.*\) \*\*(?P<stage>.*)\*\*\n" \
                       r"\n(?P<resources>(<!-- .* -->\n)+)" \
                       r"(?P<examples>\n#### examples\n" \
                       r"(\n- (?P<example_desc>.*)\n" \
                       r"    ```bash\n" \
                       r"(        (?P<example_cmd>.*)\n)+" \
                       r"    ```\n)+)?"
    VERSION_INFO_RE = re.compile(_VERSION_INFO_RE, re.MULTILINE)
    _EXAMPLE_INFO_RE = r"- (?P<example_desc>.*)\n" \
                       r"    ```bash\n" \
                       r"(?P<example_cmds>(        (.*)\n)+)" \
                       r"    ```\n"
    EXAMPLE_INFO_RE = re.compile(_EXAMPLE_INFO_RE, re.MULTILINE)
    _EXAMPLE_LINE_RE = r"        (?P<example_cmd>.*)\n"
    EXAMPLE_LINE_RE = re.compile(_EXAMPLE_LINE_RE, re.MULTILINE)

    def __init__(self, names, short_help, uri, aaz_path):
        self.names = names
        self.short_help = short_help
        self.uri = uri
        self.aaz_path = aaz_path

    def load(self):
        with open(self.aaz_path + self.uri, "r", encoding="utf-8") as f:
            content = f.read()
            command = self.parse_command_info(content, self.names)
            return command

    @classmethod
    def parse_command_info(cls, info, cmd_names):
        command_match = re.match(cls.COMMAND_INFO_RE, info)
        if not command_match:
            logger.warning(f"Invalid command info markdown: \n{info}")
            return None
        if command_match.group("lines_help"):
            lines_help = command_match.group("lines_help").strip().split("\\\n")
        else:
            lines_help = None
        help = CMDHelp()
        help.short = command_match.group("short_help").strip()
        help.lines = lines_help
        versions = []
        for version_match in re.finditer(cls.VERSION_INFO_RE, info):
            resources = []
            for resource_match in re.finditer(cls.RESOURCE_INFO_RE, version_match.group("resources")):
                resource = {
                    "plane": resource_match.group("plane"),
                    "id": resource_match.group("id"),
                    "version": resource_match.group("version"),
                }
                if resource_match.group("subresource"):
                    resource["subresource"] = resource_match.group("subresource")
                resources.append(CMDSpecsResource(raw_data=resource))
            examples = []
            for example_match in re.finditer(cls.EXAMPLE_INFO_RE, version_match.group("examples") or ''):
                example_cmd = []
                for line_match in re.finditer(cls.EXAMPLE_LINE_RE, example_match.group("example_cmds") or ''):
                    example_cmd.append(line_match.group("example_cmd"))
                example = {
                    "name": example_match.group("example_desc"),
                    "commands": example_cmd
                }
                examples.append(CMDCommandExample(raw_data=example))
            version = {
                "name": version_match.group("version_name"),
                "resources": resources,
                "examples": examples
            }
            if version_match.group("stage"):
                version["stage"] = version_match.group("stage")
            versions.append(CMDSpecsCommandVersion(raw_data=version))
        command = CMDSpecsCommand()
        command.names = cmd_names
        command.help = help
        command.versions = sorted(versions, key=lambda v: v.name)
        return command


class CMDSpecsPartialCommandTree:
    def __init__(self, aaz_path, root=None):
        self.aaz_path = aaz_path
        self._root = root or CMDSpecsPartialCommandGroup(names=["aaz"], short_help='', uri="/Commands/readme.md",
                                                         aaz_path=aaz_path).load()
        self._modified_command_groups = set()
        self._modified_commands = set()

    @property
    def root(self):
        if isinstance(self._root, CMDSpecsPartialCommandGroup):
            self._root = self._root.load()
        return self._root

    @property
    def simple_tree(self):
        """
        Build and Return a Simple Command Tree from Folder Structure
        """
        return build_simple_command_tree(self.aaz_path)

    def find_command_group(self, *cg_names):
        """
        Find command group node by names

        :param cg_names: command group names
        :return: command group node
        :rtype: CMDSpecsCommandGroup | None
        """
        node = self.root
        idx = 0
        while idx < len(cg_names):
            name = cg_names[idx]
            if not node.command_groups or name not in node.command_groups:
                return None
            node = node.command_groups[name]
            idx += 1
        return node

    def find_command(self, *cmd_names):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid command name: '{' '.join(cmd_names)}'")

        node = self.find_command_group(*cmd_names[:-1])
        if not node:
            return None
        name = cmd_names[-1]
        if not node.commands or name not in node.commands:
            return None
        leaf = node.commands[name]
        return leaf

    def iter_command_groups(self, *root_cg_names):
        root = self.find_command_group(*root_cg_names)
        if root:
            nodes = [root]
            i = 0
            while i < len(nodes):
                yield nodes[i]
                for node in (nodes[i].command_groups or {}).values():
                    nodes.append(node)
                i += 1

    def iter_commands(self, *root_node_names):
        for node in self.iter_command_groups(*root_node_names):
            for leaf in (node.commands or {}).values():
                yield leaf

    def create_command_group(self, *cg_names):
        if len(cg_names) < 1:
            raise exceptions.InvalidAPIUsage(f"Invalid Command Group name: '{' '.join(cg_names)}'")
        node = self.root
        idx = 0
        while idx < len(cg_names):
            name = cg_names[idx]
            if node.commands and name in node.commands:
                raise exceptions.InvalidAPIUsage(f"Invalid Command Group name: conflict with Command name: "
                                                 f"'{' '.join(cg_names[:idx+1])}'")
            if not node.command_groups or name not in node.command_groups:
                if not node.command_groups:
                    node.command_groups = {}
                names = [*cg_names[:idx+1]]
                node.command_groups[name] = CMDSpecsCommandGroup({
                    "names": names
                })
                self._modified_command_groups.add(cg_names[:idx+1])
            node = node.command_groups[name]
            idx += 1
        return node

    def update_command_group_by_ws(self, ws_node):
        command_group = self.create_command_group(*ws_node.names)
        if ws_node.help:
            if not command_group.help:
                command_group.help = CMDHelp()
            if ws_node.help.short:
                command_group.help.short = ws_node.help.short
            if ws_node.help.lines:
                command_group.help.lines = [*ws_node.help.lines]
        self._modified_command_groups.add(tuple([*ws_node.names]))
        return command_group

    def delete_command_group(self, *cg_names):
        for _ in self.iter_commands(*cg_names):
            raise exceptions.ResourceConflict("Cannot delete command group with commands")
        parent = self.find_command_group(*cg_names[:-1])
        name = cg_names[-1]
        if not parent or not parent.command_groups or name not in parent.command_groups:
            return False
        del parent.command_groups[name]
        if not parent.command_groups:
            parent.command_groups = None

        self._modified_command_groups.add(cg_names)

        if not parent.command_groups and not parent.commands:
            # delete empty parent command group
            self.delete_command_group(*cg_names[:-1])
        return True

    def create_command(self, *cmd_names):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid Command name: '{' '.join(cmd_names)}'")
        node = self.create_command_group(*cmd_names[:-1])
        name = cmd_names[-1]
        if node.command_groups and name in node.command_groups:
            raise exceptions.InvalidAPIUsage(f"Invalid Command name: conflict with Command Group name: "
                                             f"'{' '.join(cmd_names)}'")
        if not node.commands:
            node.commands = {}
        elif name in node.commands:
            return node.commands[name]

        command = CMDSpecsCommand()
        command.names = list(cmd_names)
        node.commands[name] = command
        self._modified_commands.add(cmd_names)

        return command

    def delete_command(self, *cmd_names):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid Command name: '{' '.join(cmd_names)}'")
        parent = self.find_command_group(*cmd_names[:-1])
        name = cmd_names[-1]
        if not parent or not parent.commands or name not in parent.commands:
            return False
        command = parent.commands[name]
        if command.versions:
            raise exceptions.ResourceConflict("Cannot delete command with versions")
        del parent.commands[name]
        if not parent.commands:
            parent.commands = None

        self._modified_commands.add(cmd_names)

        if not parent.command_groups and not parent.commands:
            # delete empty parent command group
            self.delete_command_group(*cmd_names[:-1])
        return True

    def delete_command_version(self, *cmd_names, version):
        if len(cmd_names) < 2:
            raise exceptions.InvalidAPIUsage(f"Invalid Command name: '{' '.join(cmd_names)}'")
        command = self.find_command(*cmd_names)
        if not command or not command.versions:
            return False
        match_idx = None
        for idx, v in enumerate(command.versions):
            if v.name == version:
                match_idx = idx
                break
        if match_idx is None:
            return False

        command.versions = command.versions[:match_idx] + command.versions[match_idx+1:]

        self._modified_commands.add(cmd_names)

        if not command.versions:
            # delete empty command
            self.delete_command(*cmd_names)
        return True

    def update_command_version(self, *cmd_names, plane, cfg_cmd):
        command = self.create_command(*cmd_names)

        version = None
        for v in (command.versions or []):
            if v.name == cfg_cmd.version:
                version = v
                break

        if not version:
            version = CMDSpecsCommandVersion()
            version.name = cfg_cmd.version
            if not command.versions:
                command.versions = []
            command.versions.append(version)

        # update version resources
        version.resources = []
        for r in cfg_cmd.resources:
            resource = CMDSpecsResource()
            resource.plane = plane
            resource.id = r.id
            resource.version = r.version
            resource.subresource = r.subresource
            version.resources.append(resource)

        self._modified_commands.add(cmd_names)

    def update_command_by_ws(self, ws_leaf):
        command = self.find_command(*ws_leaf.names)
        if not command:
            # make sure the command exist, if command not exist, then run update_resource_cfg first
            raise exceptions.InvalidAPIUsage(f"Command isn't exist: '{' '.join(ws_leaf.names)}'")

        cmd_version = None
        for v in (command.versions or []):
            if v.name == ws_leaf.version:
                cmd_version = v
                break
        if not cmd_version:
            raise exceptions.InvalidAPIUsage(f"Command in version isn't exist: "
                                             f"'{' '.join(ws_leaf.names)}' '{ws_leaf.version}'")

        # compare resources
        leaf_resources = {(r.id, r.version) for r in ws_leaf.resources}
        cmd_version_resources = {(r.id, r.version) for r in cmd_version.resources}
        if leaf_resources != cmd_version_resources:
            raise exceptions.InvalidAPIUsage(f"The resources in version don't match the resources of workspace leaf: "
                                             f"{leaf_resources} != {cmd_version_resources}")

        # update stage
        cmd_version.stage = ws_leaf.stage

        # update examples
        if ws_leaf.examples:
            cmd_version.examples = [CMDCommandExample(e.to_primitive()) for e in ws_leaf.examples]

        # update help
        if ws_leaf.help:
            if not command.help:
                command.help = CMDHelp()
            if ws_leaf.help.short:
                command.help.short = ws_leaf.help.short
            if ws_leaf.help.lines:
                command.help.lines = [*ws_leaf.help.lines]

        self._modified_commands.add(tuple(command.names))
        return command

    def verify_command_tree(self):
        details = {}
        for group in self.iter_command_groups():
            if group == self.root:
                continue
            if not group.help or not group.help.short:
                details[' '.join(group.names)] = {
                    'type': 'group',
                    'help': "Miss short summary."
                }

        for cmd in self.iter_commands():
            if not cmd.help or not cmd.help.short:
                details[' '.join(cmd.names)] = {
                    'type': 'command',
                    'help': "Miss short summary."
                }
        if details:
            raise exceptions.VerificationError(message="Invalid Command Tree", details=details)

    def verify_updated_command_tree(self):
        details = {}
        for group_names in self._modified_command_groups:
            group = self.find_command_group(*group_names)
            if group == self.root:
                continue
            if not group:
                details[' '.join(group_names)] = {
                    'type': 'group',
                    'help': "Group not found."
                }
            elif not group.help or not group.help.short:
                details[' '.join(group.names)] = {
                    'type': 'group',
                    'help': "Miss short summary."
                }

        for cmd_names in self._modified_commands:
            cmd = self.find_command(*cmd_names)
            if not cmd:
                details[' '.join(cmd_names)] = {
                    'type': 'command',
                    'help': "Command not found."
                }
            elif not cmd.help or not cmd.help.short:
                details[' '.join(cmd.names)] = {
                    'type': 'command',
                    'help': "Miss short summary."
                }
        if details:
            raise exceptions.VerificationError(message="Invalid Command Tree", details=details)

    def to_model(self):
        tree = CMDSpecsCommandTree()
        tree.root = self.root
        return tree



from schematics.models import Model
from schematics.types import StringType, ModelType, UTCDateTimeType, ListType, DictType
from command.model.configuration._fields import CMDCommandNameField, CMDVersionField
from command.model.configuration import CMDStageField, CMDHelp, CMDResource, CMDCommandExample
from utils.fields import PlaneField


class CMDCommandTreeLeaf(Model):
    names = ListType(field=CMDCommandNameField(), min_size=1,
                     required=True)   # full name of a command
    stage = CMDStageField()
    version = CMDVersionField(required=True)

    help = ModelType(CMDHelp)
    # the azure resources used in this command
    resources = ListType(ModelType(CMDResource), min_size=1)
    examples = ListType(ModelType(CMDCommandExample))

    class Options:
        serialize_when_none = False


class CMDCommandTreeNode(Model):
    names = ListType(field=CMDCommandNameField(), min_size=1,
                     required=True)   # full name of a command group
    stage = CMDStageField()

    help = ModelType(CMDHelp)
    command_groups = DictType(
        field=ModelType("CMDCommandTreeNode"),
        serialized_name='commandGroups',
        deserialize_from='commandGroups'
    )
    commands = DictType(
        field=ModelType(CMDCommandTreeLeaf)
    )

    class Options:
        serialize_when_none = False


class CMDEditorWorkspace(Model):
    # this property updated when workspace saved in file.
    version = UTCDateTimeType(required=True)
    name = StringType(required=True)
    plane = PlaneField(required=True)
    mod_names = StringType(
        serialized_name='modNames',
        deserialize_from='modNames',
    )  # The mod names in azure-rest-api-specs. The names are joined by '/'
    resource_provider = StringType(
        serialized_name='resourceProvider',
        deserialize_from='resourceProvider'
    )
    source = StringType(
        choices=['OpenAPI', 'TypeSpec'],
        default='OpenAPI',
    )  # The type of the api spec source. It can be 'OpenAPI' or 'TypeSpec'
    command_tree = ModelType(
        CMDCommandTreeNode,
        required=True,
        serialized_name='commandTree',
        deserialize_from='commandTree'
    )   # the root node

    class Options:
        serialize_when_none = False



import click
import json
import logging
import os
import re
import sys
from collections import defaultdict
from flask import Blueprint

from command.controller.specs_manager import AAZSpecsManager
from command.controller.workspace_manager import WorkspaceManager
from command.model.configuration import CMDHelp
from swagger.controller.specs_manager import SwaggerSpecsManager
from swagger.model.specs import TypeSpecResourceProvider
from swagger.utils.source import SourceTypeEnum
from swagger.utils.tools import swagger_resource_path_to_resource_id

from utils.config import Config
from utils.exceptions import InvalidAPIUsage

logger = logging.getLogger('aaz')

bp = Blueprint('aaz-cmds', __name__, url_prefix='/AAZ/CMDs', cli_group="command-model")
bp.cli.short_help = "Manage command models in aaz."


def path_type(ctx, param, value):
    import os
    return os.path.expanduser(value)


def resource_id_type(value):
    return swagger_resource_path_to_resource_id(value)


@bp.cli.command("generate-from-swagger", short_help="Generate command models into aaz from swagger specs")
@click.option(
    "--swagger-path", '-s',
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_PATH,
    callback=Config.validate_and_setup_swagger_path,
    expose_value=False,
    help="The local path of azure-rest-api-specs repo. Official repo is https://github.com/Azure/azure-rest-api-specs"
)
@click.option(
    "--swagger-module-path", "--sm",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_MODULE_PATH,
    callback=Config.validate_and_setup_swagger_module_path,
    expose_value=False,
    help="The local path of swagger in module level. It can be substituted for --swagger-path."
)
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--module", '-m',
    default=Config.DEFAULT_SWAGGER_MODULE,
    required=not Config.DEFAULT_SWAGGER_MODULE,
    callback=Config.validate_and_setup_default_swagger_module,
    expose_value=False,
    help="The name of swagger module."
)
@click.option(
    "--resource-provider", "--rp",
    default=Config.DEFAULT_RESOURCE_PROVIDER,
    required=not Config.DEFAULT_RESOURCE_PROVIDER,
    callback=Config.validate_and_setup_default_resource_provider,
    expose_value=False,
    help="The resource provider name."
)
@click.option(
    "--swagger-tag", "--tag",
    required=True,
    help="Swagger tag with input files."
)
@click.option(
    "--workspace-path",
    help="The path to export the workspace for modification."
)
def generate_command_models_from_swagger(swagger_tag, workspace_path=None):
    try:
        swagger_specs = SwaggerSpecsManager()
        aaz_specs = AAZSpecsManager()

        module_manager = swagger_specs.get_module_manager(Config.DEFAULT_PLANE, Config.DEFAULT_SWAGGER_MODULE)
        rp = module_manager.get_openapi_resource_provider(Config.DEFAULT_RESOURCE_PROVIDER)

        resource_map = rp.get_resource_map_by_tag(swagger_tag)
        if not resource_map:
            raise InvalidAPIUsage(f"Tag `{swagger_tag}` is not exist")

        version_resource_map = {}
        for resource_id, version_map in resource_map.items():
            v_list = [v for v in version_map]
            if len(v_list) > 1:
                raise InvalidAPIUsage(f"Tag `{swagger_tag}` contains multiple api versions of one resource", payload={
                    "Resource": resource_id,
                    "versions": v_list,
                })
            v = v_list[0]
            if v not in version_resource_map:
                version_resource_map[v] = []
            version_resource_map[v].append({
                "id": resource_id
            })

        mod_names = Config.DEFAULT_SWAGGER_MODULE
        ws = WorkspaceManager.new(
            name=Config.DEFAULT_SWAGGER_MODULE,
            plane=Config.DEFAULT_PLANE,
            folder=workspace_path or WorkspaceManager.IN_MEMORY,  # if workspace path exist, use workspace else use in memory folder
            mod_names=mod_names,
            resource_provider=rp.name,
            swagger_manager=swagger_specs,
            aaz_manager=aaz_specs,
            source=SourceTypeEnum.OpenAPI,
        )
        for version, resources in version_resource_map.items():
            ws.add_new_resources_by_swagger(
                mod_names=mod_names, version=version, resources=resources
            )

        # provide default short summary
        for node in ws.iter_command_tree_nodes():
            if not node.help:
                node.help = CMDHelp()
            if not node.help.short:
                node.help.short = f"Manage {node.names[-1]}"

        for leaf in ws.iter_command_tree_leaves():
            if not leaf.help:
                leaf.help = CMDHelp()
            if not leaf.help.short:
                n = leaf.names[-1]
                n = n[0].upper() + n[1:]
                leaf.help.short = f"{n} {leaf.names[-2]}"
            # generate examples
            cfg_editor = ws.load_cfg_editor_by_command(leaf)
            command = cfg_editor.find_command(*leaf.names)
            examples = ws.generate_examples_by_swagger(leaf, command)
            leaf.examples = examples

        if not ws.is_in_memory:
            ws.save()

        ws.generate_to_aaz()

    except InvalidAPIUsage as err:
        logger.error(err)
        sys.exit(1)
    except ValueError as err:
        logger.error(err)
        sys.exit(1)


@bp.cli.command("generate-all", short_help="Fully generate data model from OpenAPI specification, mainly for use in https://github.com/magodo/az-rs.")
@click.option(
    "--swagger-path", '-s',
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_PATH,
    callback=Config.validate_and_setup_swagger_path,
    expose_value=False,
    help="The local path of azure-rest-api-specs repo. Official repo is https://github.com/Azure/azure-rest-api-specs"
)
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--module", '-m',
    default=Config.DEFAULT_SWAGGER_MODULE,
    callback=Config.validate_and_setup_default_swagger_module,
    expose_value=False,
    help="The name of swagger module."
)
def generate_all():
    def update_strategy(resource):
        has_patch = "patch" in resource.operations.values()
        # has_generic_update = {"get", "put"}.issubset(resource.operations.values())
        if has_patch:
            return "PatchOnly"
        # if has_generic_update:  # fallback to get + put
        #     return "None"
        return None

    def normalize_resource_map(resource_map):
        version_resource_map = defaultdict(list)
        for resource_id, version_map in resource_map.items():
            for version in sorted(version_map.keys())[-4:]:  # only care about the latest 4 versions
                resource = {"id": resource_id}
                update_by = update_strategy(version_map[version])
                if update_by:
                    resource["options"] = {"update_by": update_by}

                version_resource_map[version].append(resource)

        return version_resource_map

    def to_aaz(module_name):
        module_manager = SwaggerSpecsManager().get_module_manager(Config.DEFAULT_PLANE, module_name)
        rps = module_manager.get_resource_providers()

        for r in rps:
            if isinstance(r, TypeSpecResourceProvider):
                continue

            rp = module_manager.get_openapi_resource_provider(r.name)
            version_resource_map = normalize_resource_map(rp.get_resource_map())

            for version, resources in version_resource_map.items():
                ws = WorkspaceManager.new(
                    name=module_name,
                    plane=Config.DEFAULT_PLANE,
                    folder=WorkspaceManager.IN_MEMORY,
                    mod_names=module_name,
                    resource_provider=rp.name,
                    swagger_manager=SwaggerSpecsManager(),
                    aaz_manager=AAZSpecsManager(),
                    source=SourceTypeEnum.OpenAPI,
                )
                ws.add_new_resources_by_swagger(mod_names=module_name, version=version, resources=resources)

                # provide default short summary
                for node in ws.iter_command_tree_nodes():
                    if not node.help:
                        node.help = CMDHelp()
                    if not node.help.short:
                        node.help.short = f"Manage {node.names[-1]}."

                for leaf in ws.iter_command_tree_leaves():
                    if not leaf.help:
                        leaf.help = CMDHelp()
                    if not leaf.help.short:
                        n = leaf.names[-1]
                        n = n[0].upper() + n[1:]
                        leaf.help.short = f"{n} {leaf.names[-2]}."

                ws.generate_to_aaz()

    try:
        if Config.DEFAULT_SWAGGER_MODULE:
            to_aaz(Config.DEFAULT_SWAGGER_MODULE)

        else:
            spec_path = os.path.join(Config.SWAGGER_PATH, "specification")
            for item in os.listdir(spec_path):
                curr_path = os.path.join(spec_path, item)
                if os.path.isfile(curr_path):
                    continue

                to_aaz(item)

    except (InvalidAPIUsage, ValueError) as err:
        logger.error(err, exc_info=True)
        raise sys.exit(1)


@bp.cli.command("verify", short_help="Verify data consistency within `aaz` repository.")
@click.option(
    "--aaz-path", "-a",
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="Path of `aaz` repository."
)
@click.option(
    "--target", "-t",
    help="Target module to focus on."
)
def verify(target):
    def verify_resource(model, path):
        if "commandGroups" not in model:
            return

        for grp in model["commandGroups"]:
            base_path = os.path.join(path, *grp["name"].split())
            if not os.path.exists(base_path):
                raise FileNotFoundError(base_path)

            for cmd in grp.get("commands", []):
                file_path = os.path.join(base_path, f"_{cmd['name']}.md")
                if not os.path.isfile(file_path):
                    raise FileNotFoundError(file_path)

            verify_resource(grp, base_path)

    def verify_command(file_path):
        with open(file_path, "r", encoding="utf-8") as fp:
            content = fp.read()

        base_path = os.path.dirname(file_path)
        curr_grp = " ".join(os.path.relpath(base_path, aaz.commands_folder).split(os.sep))
        curr_cmd = os.path.splitext(os.path.basename(file_path))[0][1:]

        paths = re.findall(r"]\(([^)]+)\)", content)
        for path in paths:
            json_path = os.path.join(Config.AAZ_PATH, os.path.splitext(path)[0][1:] + ".json")
            json_path = os.path.normpath(json_path)

            if not os.path.exists(json_path):
                raise Exception(f"{json_path} defined in {file_path} is missing.")

            with open(json_path, "r", encoding="utf-8", errors="ignore") as fp:
                model = json.load(fp)

            try:
                verify_resource(model, aaz.commands_folder)

            except FileNotFoundError as e:
                raise Exception(f"Cannot find {e} defined in {json_path}.")

            target = curr_grp
            while target:
                try:
                    for grp in model["commandGroups"]:
                        if target.startswith(grp["name"]):
                            target = target[len(grp["name"]):].strip()
                            model = grp

                            break

                except KeyError:
                    raise Exception(f"{curr_grp} has no corresponding definition in {json_path}.")

            commands = model["commands"]
            if not any(cmd["name"] == curr_cmd for cmd in commands):
                raise Exception(f"There is no {curr_cmd} command info in {json_path}.")

            model_set.add(json_path)

    model_set = set()
    aaz = AAZSpecsManager()

    parent = aaz.commands_folder
    stack = [os.path.join(parent, target)] if target else [
        os.path.join(parent, i)
        for i in os.listdir(parent)
        if os.path.isdir(os.path.join(parent, i))
    ]

    while stack:
        curr_path = stack.pop()
        if os.path.isdir(curr_path):
            readme_path = os.path.join(curr_path, "readme.md")
            if not os.path.exists(readme_path):
                raise Exception(f"Missing `readme.md` under {curr_path}.")

            with open(readme_path, "r", encoding="utf-8") as fp:
                content = fp.read()

            matches = re.findall(r"## (.+)\n\n(((?!\n##)[\s\S])+)", content)
            for match in matches:
                level = match[0]
                items = re.findall(r"- \[([^[\]]+)]", match[1])

                if level == "Commands":
                    if len(items) != len(set(items)):
                        raise Exception(f"{readme_path} has duplicate command names.")

                    items = set(items)

                    files = {i for i in os.listdir(curr_path) if os.path.isfile(os.path.join(curr_path, i))}
                    files.remove("readme.md")

                    if (cmd_set := set(map(lambda x: x[1:-3], files))) != items:  # _<command_name>.md
                        diff = cmd_set - items or items - cmd_set
                        raise Exception(f"Command info {diff} doesn't match in {readme_path}.")

                    for file in files:
                        verify_command(os.path.join(curr_path, file))

                else:
                    if len(items) != len(set(items)):
                        raise Exception(f"{readme_path} has duplicate command group names.")

                    items = set(items)

                    folders = {i for i in os.listdir(curr_path) if os.path.isdir(os.path.join(curr_path, i))}
                    if folders != items:
                        diff = folders - items or items - folders
                        raise Exception(f"Command group info {diff} doesn't match in {readme_path}.")

                    for folder in folders:
                        stack.append(os.path.join(curr_path, folder))

    if not target:
        for root, _, files in os.walk(aaz.resources_folder):
            for file in files:
                if not file.endswith(".json") or file.startswith("client"):  # support data-plane
                    continue

                file_path = os.path.join(root, file)
                if file_path not in model_set:
                    raise Exception(f"{file_path} is redundant.")


@bp.cli.command("transform", short_help="Transform into machine-friendly metadata.")
@click.option(
    "--aaz-path", "-a",
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="Path of `aaz` repository."
)
def transform():
    def nested_dict():
        return defaultdict(nested_dict)

    def process_resource(model):
        resources = []
        for resource in model["resources"]:
            resources.append({
                "id": resource["id"],
                "plane": model["plane"]
            })

        return resources

    def export_command(xml_path, file_name):
        json_path = os.path.normpath(os.path.join(Config.AAZ_PATH, f"{os.path.splitext(xml_path)[0][1:]}.json"))
        with open(json_path, "r", encoding="utf-8", errors="ignore") as f:
            model = json.load(f)

        target = file_name.split("_")[-2]  # command name

        data = {"resources": process_resource(model)}
        for cmd in model["commandGroups"][0].get("commands", []):
            if cmd["name"] == target:
                other = {k: v for k, v in cmd.items() if k not in {"name", "version", "resources"}}
                data = {**data, **other}

                break

        with open(os.path.join("./metadata/commands", file_name), "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, separators=(",", ":"), sort_keys=True)

    def build_command(path):
        command = defaultdict(nested_dict)
        with open(path, "r", encoding="utf-8") as f:
            content = f.read()

        command["help"]["short"] = re.findall(r"# \[Command\].*?\n\n(.*?)\n", content)[0]
        command["versions"] = []

        matches = re.findall(r"### \[(.*?)\]\((.*?)\)", content)
        for version, xml_path in matches:
            command["versions"].insert(0, version)

            *grp, cmd_md = os.path.relpath(path, aaz_commands).split(os.sep)
            export_command(xml_path, f"{'_'.join(grp)}_{cmd_md[1:-3]}_{version}.json")

        return command

    def build_group(path):
        group = defaultdict(nested_dict)
        for item in sorted(os.listdir(path)):
            curr_path = os.path.join(path, item)
            if item == "readme.md":
                with open(curr_path, "r", encoding="utf-8") as f:
                    content = f.read()

                group["help"]["short"] = re.findall(r"# \[Group\].*?\n\n(.*?)\n", content)[0]
                continue

            if os.path.isdir(curr_path):
                group["commandGroups"][item] = build_group(curr_path)

            else:
                group["commands"][item[1:-3]] = build_command(curr_path)

        return group

    os.makedirs("metadata/commands")
    index = defaultdict(nested_dict)

    aaz_commands = AAZSpecsManager().commands_folder
    for item in sorted(os.listdir(aaz_commands)):
        curr_path = os.path.join(aaz_commands, item)
        if os.path.isdir(curr_path):
            index[item] = build_group(curr_path)

    with open(os.path.join("./metadata", "index.json"), "w", encoding="utf-8") as f:
        json.dump(index, f, ensure_ascii=False, separators=(",", ":"))



from base64 import urlsafe_b64encode, urlsafe_b64decode, b64encode, b64decode


def b64encode_str(value, urlsafe=True):
    if urlsafe:
        return urlsafe_b64encode(value.encode()).decode()
    else:
        return b64encode(value.encode()).decode()


def b64decode_str(value, urlsafe=True):
    if urlsafe:
        return urlsafe_b64decode(value).decode()
    else:
        return b64decode(value).decode()


__all__ = ['b64encode_str', 'b64decode_str']



from cli.model.common._fields import CLIClientConfigField
from schematics.models import Model
from schematics.types import StringType
from utils.fields import PlaneField


class CLIAtomicClient(Model):

    plane = PlaneField(required=True)
    name = StringType(required=True)  # client name registered in azure-cli-core, it will include the plane name and currently module name
    registered_name = StringType(
        required=True,
        serialized_name="registeredName",
        deserialize_from="registeredName"
    )
    cls_name = StringType(
        required=True,
        serialized_name="clsName",
        deserialize_from="clsName",
    )
    cfg = CLIClientConfigField()

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import StringType, ModelType, BooleanType


class XmsLongRunningOperationField(BooleanType):
    """
    Some requests like creating/deleting a resource cannot be carried out immediately. In such a situation, the server sends a 201 (Created) or 202 (Accepted) and provides a link to monitor the status of the request. When such an operation is marked with extension "x-ms-long-running-operation": true, in OpenAPI, the generated code will know how to fetch the link to monitor the status. It will keep on polling at regular intervals till the request reaches one of the terminal states: Succeeded, Failed, or Canceled.

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-long-running-operation
    """

    def __init__(self, **kwargs):
        super(XmsLongRunningOperationField, self).__init__(
            serialized_name='x-ms-long-running-operation',
            deserialize_from='x-ms-long-running-operation',
            **kwargs
        )


class XmsLongRunningOperationOptions(Model):
    """
    When x-ms-long-running-operation is specified, there should also be a x-ms-long-running-operation-options specified.

    See Azure RPC Spec for asynchronous operation notes.

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-long-running-operation-options

    It will keep on polling at regular intervals till the request reaches one of the terminal states: Succeeded, Failed, or Canceled.
    """

    final_state_via = StringType(
        choices=(
            'azure-async-operation',  # (default if not specified) poll until terminal state, the final response will be available at the uri pointed to by the header Azure-AsyncOperation
            'location',  # poll until terminal state, the final response will be available at the uri pointed to by the header Location
            'original-uri'  # poll until terminal state, the final response will be available via GET at the original resource URI. Very common for PUT operations.
        ), default='azure-async-operation',
        serialized_name='final-state-via',
        deserialize_from='final-state-via',
    )

    final_state_schema = StringType(
        serialized_name="final-state-schema",
        deserialize_from="final-state-schema",
    )


class XmsLongRunningOperationOptionsField(ModelType):

    def __init__(self, **kwargs):
        super(XmsLongRunningOperationOptionsField, self).__init__(
            XmsLongRunningOperationOptions,
            serialized_name='x-ms-long-running-operation-options',
            deserialize_from='x-ms-long-running-operation-options',
            **kwargs
        )



import logging
import os
import json
from datetime import datetime

from .client_cfg_reader import ClientCfgReader
from .workspace_helper import ArgumentUpdateMixin
from command.model.configuration import (CMDClientConfig, CMDDiffLevelEnum, CMDClientEndpointsByTemplate,
                                         CMDClientEndpointsByHttpOperation, CMDClientEndpoints,
                                         CMDClientEndpointCloudMetadataTemplate)

logger = logging.getLogger('aaz')


class WorkspaceClientCfgEditor(ClientCfgReader, ArgumentUpdateMixin):
    
    @staticmethod
    def get_cfg_path(ws_folder):
        path = os.path.join(ws_folder, 'client.json')
        return path

    @classmethod
    def load_client_cfg(cls, ws_folder):
        path = cls.get_cfg_path(ws_folder)
        if not os.path.exists(path):
            return None
        with open(path, 'r') as f:
            data = json.load(f)
        cfg = CMDClientConfig(data)
        cfg_editor = cls(cfg)
        cfg_editor.reformat()
        return cfg_editor

    @classmethod
    def new_client_cfg(cls, plane, auth, endpoints, ref_cfg: ClientCfgReader = None):
        assert isinstance(endpoints, CMDClientEndpoints)
        cfg = CMDClientConfig(raw_data={
            "plane": plane,
            "auth": auth
        })
        cfg.endpoints = endpoints
        ref_args = ref_cfg.cfg.arg_group.args if ref_cfg and ref_cfg.cfg.arg_group else None
        cfg.generate_args(ref_args=ref_args)
        if not ref_cfg or cfg.endpoints.diff(ref_cfg.cfg.endpoints, CMDDiffLevelEnum.Structure) or cfg.auth.diff(ref_cfg.cfg.auth, CMDDiffLevelEnum.Structure):
            # when endpoints or auth changed then bump up version
            cfg.version = datetime.utcnow()
        else:
            cfg.version = ref_cfg.cfg.version

        cfg_editor = cls(cfg)
        cfg_editor.reformat()
        return cfg_editor

    @classmethod
    def new_client_endpoints_by_template(cls, templates, cloud_medadata):
        endpoints = CMDClientEndpointsByTemplate(raw_data={
            "templates": templates,
        })
        if cloud_medadata:
            endpoints.cloud_metadata = CMDClientEndpointCloudMetadataTemplate(raw_data=cloud_medadata)
        endpoints.prepare()
        return endpoints

    @classmethod
    def new_client_endpoints_by_http_operation(cls, resource, operation, selector):
        endpoints = CMDClientEndpointsByHttpOperation(raw_data={
            "resource": resource,
            "selector": selector,
            "operation": operation,
        })
        endpoints.prepare()
        return endpoints

    def __init__(self, cfg):
        super().__init__(cfg)

    def reformat(self):
        self.cfg.reformat()

    def update_arg_by_var(self, arg_var, **kwargs):
        arg = self.find_arg_by_var(arg_var=arg_var)
        if not arg:
            return None
        self._update_arg(arg, **kwargs)
        self.cfg.version = datetime.utcnow()
        self.reformat()



import logging

from schematics.models import Model
from schematics.types import ModelType, ListType, PolyModelType, StringType

from ._arg_group import CMDArgGroup
from ._arg import CMDClsArgBase
from ._condition import CMDCondition
from ._fields import CMDDescriptionField, CMDVersionField, CMDCommandNameField, CMDBooleanField, CMDConfirmation
from ._operation import CMDOperation, CMDHttpOperation, CMDInstanceDeleteOperation
from ._http_response_body import CMDHttpResponseJsonBody
from ._schema import CMDClsSchemaBase, CMDArraySchemaBase, CMDStringSchemaBase, CMDObjectSchemaBase
from ._output import CMDOutput, CMDArrayOutput, CMDObjectOutput, CMDStringOutput
from ._resource import CMDResource
from ._utils import CMDArgBuildPrefix, CMDDiffLevelEnum
from ._subresource_selector import CMDSubresourceSelector, CMDJsonSubresourceSelector
from ._selector_index import CMDArrayIndexBase, CMDObjectIndexBase, CMDObjectIndexDiscriminator, CMDObjectIndexAdditionalProperties
from utils import exceptions

logger = logging.getLogger('aaz')


class CMDCommand(Model):
    # properties as tags
    name = CMDCommandNameField(required=True)
    version = CMDVersionField(required=True)

    description = CMDDescriptionField()

    # properties as nodes
    resources = ListType(ModelType(CMDResource), min_size=1)  # the azure resources used in this command
    arg_groups = ListType(
        ModelType(CMDArgGroup),
        serialized_name='argGroups',
        deserialize_from='argGroups',
    )
    conditions = ListType(ModelType(CMDCondition))
    subresource_selector = PolyModelType(
        CMDSubresourceSelector,
        allow_subclasses=True,
        serialized_name="subresourceSelector",
        deserialize_from="subresourceSelector"
    )
    operations = ListType(PolyModelType(CMDOperation, allow_subclasses=True), min_size=1)
    outputs = ListType(PolyModelType(CMDOutput, allow_subclasses=True), min_size=1)  # support to add outputs in different formats, such table

    confirmation = CMDConfirmation()  # support to prompt for confirmation - optional

    class Options:
        serialize_when_none = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.arg_cls_register_map = None
        self.schema_cls_register_map = None

    def generate_args(self, ref_args=None, ref_options=None):
        if not ref_args:
            ref_args = []
            if self.arg_groups:
                for group in self.arg_groups:
                    ref_args.extend(group.args)
            ref_args = ref_args or None

        arguments = {}
        has_subresource = False
        if self.subresource_selector:
            has_subresource = True
            for arg in self.subresource_selector.generate_args(ref_args=ref_args):
                if arg.var not in arguments:
                    if ref_options and arg.var in ref_options:
                        # replace generated options by ref_options
                        arg.options = [*ref_options[arg.var]]
                    arguments[arg.var] = arg

        for op in self.operations:
            for arg in op.generate_args(ref_args=ref_args, has_subresource=has_subresource):
                if arg.var not in arguments:
                    if ref_options and arg.var in ref_options:
                        # replace generated options by ref_options
                        arg.options = [*ref_options[arg.var]]
                    arguments[arg.var] = arg

        arguments = handle_duplicated_options(
            arguments, has_subresource=has_subresource, operation_id=self.operations[-1].operation_id)
        self.arg_groups = self._build_arg_groups(arguments)

    def generate_outputs(self, ref_outputs=None, pageable=None):
        if not ref_outputs:
            if self.outputs:
                ref_outputs = [*self.outputs]

        client_flatten = True
        if ref_outputs and isinstance(ref_outputs[0], (CMDArrayOutput, CMDObjectOutput)):
            client_flatten = ref_outputs[0].client_flatten

        output = None
        if self.subresource_selector:
            delete_op = False
            for op in self.operations:
                if isinstance(op, CMDInstanceDeleteOperation):
                    delete_op = True
            # instance delete action does not need output
            if not delete_op:
                output = self._build_output_type_by_subresource_selector(self.subresource_selector)
                output.ref = self.subresource_selector.var
                output.client_flatten = client_flatten
        else:
            for op in self.operations:
                if isinstance(op, CMDHttpOperation):
                    op_output = self.build_output_by_operation(op, pageable, client_flatten)
                    if op_output:
                        output = op_output
            if output and ref_outputs:
                assert len(ref_outputs) == 1, "Only support one reference output"
                ref_output = ref_outputs[0]
                if isinstance(ref_output, CMDArrayOutput) and ref_output.ref.startswith(output.ref + '.'):
                    # inherit pageable
                    output = CMDArrayOutput()
                    output.ref = ref_output.ref
                    output.next_link = ref_output.next_link
                    output.client_flatten = client_flatten

        self.outputs = [output] if output else None

    @classmethod
    def build_output_by_operation(cls, op, pageable=None, client_flatten=True):
        assert isinstance(op, CMDHttpOperation)
        output = None
        for resp in op.http.responses:
            if resp.is_error:
                continue
            if resp.body is None:
                if 204 in resp.status_codes and op.http.request.method == "delete":
                    # no content response for delete operation
                    return None
                continue
            if isinstance(resp.body, CMDHttpResponseJsonBody):
                body_json = resp.body.json
                if pageable and pageable.item_name:
                    output = CMDArrayOutput()
                    output.ref = f"{body_json.var}.{pageable.item_name}"
                    if pageable.next_link_name:
                        output.next_link = f"{body_json.var}.{pageable.next_link_name}"
                else:
                    output = cls._build_output_type_schema(body_json.schema)
                    output.ref = body_json.var
                output.client_flatten = client_flatten
            else:
                raise NotImplementedError()
        return output

    def reformat(self, **kwargs):
        self.resources = sorted(self.resources, key=lambda r: r.id)
        try:
            self._reformat_arg_groups(**kwargs)
            self._reformat_operations(**kwargs)
        except exceptions.VerificationError as err:
            err.payload['details'] = {
                "type": "Command",
                "name": self.name,
                "resources": [resource.to_primitive() for resource in self.resources],
                "details": err.payload['details']
            }
            raise err

    def _reformat_arg_groups(self, **kwargs):
        if not self.arg_groups:
            return

        used_options = set()
        args_in_group = {}
        for arg_group in self.arg_groups:
            for arg in arg_group.args:
                group_name = arg.group or ""
                if group_name not in args_in_group:
                    args_in_group[group_name] = []
                args_in_group[group_name].append(arg)
                for option in arg.options:
                    if option in used_options:
                        raise exceptions.VerificationError(
                            message=f"Argument option '{option}' duplicated.",
                            details={
                                "type": "Argument",
                                "options": arg.options,
                                "var": arg.var,
                            }
                        )
                    used_options.add(option)
        arg_groups = []
        for group_name, args in args_in_group.items():
            arg_group = CMDArgGroup()
            arg_group.name = group_name
            arg_group.args = args
            arg_group.reformat(**kwargs)
            arg_groups.append(arg_group)
        self.arg_groups = sorted(arg_groups, key=lambda a: a.name)

    def _reformat_operations(self, **kwargs):
        if not self.operations:
            return
        schema_cls_map = {}
        for operation in self.operations:
            operation.reformat(schema_cls_map=schema_cls_map, **kwargs)

        for key, value in schema_cls_map.items():
            if value is None:
                raise exceptions.VerificationError(
                    message=f"ReformatError: Schema Class '{key}' not defined.",
                    details=None
                )

    def link(self):
        self.arg_cls_register_map = {}
        self.schema_cls_register_map = {}

        if self.arg_groups:
            arg_cls_register_map = {}
            for arg_group in self.arg_groups:
                arg_group.register_cls(cls_register_map=arg_cls_register_map)

            for key, value in arg_cls_register_map.items():
                implement = value["implement"]
                refers = value["refers"]
                if implement is None:
                    raise exceptions.VerificationError(
                        message=f"LinkError: Argument Class '{key}' not defined.",
                        details=None
                    )
                for refer in refers:
                    assert isinstance(refer, CMDClsArgBase)
                    refer.implement = implement

            self.arg_cls_register_map = arg_cls_register_map

        if self.operations:
            schema_cls_register_map = {}
            for operation in self.operations:
                operation.register_cls(cls_register_map=schema_cls_register_map)
            for key, value in schema_cls_register_map.items():
                implement = value["implement"]
                refers = value["refers"]
                if implement is None:
                    raise exceptions.VerificationError(
                        message=f"LinkError: Schema Class '{key}' not defined.",
                        details=None
                    )
                for refer in refers:
                    assert isinstance(refer, CMDClsSchemaBase)
                    refer.implement = implement

            self.schema_cls_register_map = schema_cls_register_map

    @staticmethod
    def _build_arg_groups(arguments):
        # build argument groups
        arg_groups = {}
        for arg in arguments:
            group_name = arg.group or ""
            if group_name not in arg_groups:
                arg_groups[group_name] = {}
            if arg.var not in arg_groups[group_name]:
                arg_groups[group_name][arg.var] = arg

        groups = []
        for group_name, args in arg_groups.items():
            group = CMDArgGroup()
            group.name = group_name
            group.args = [arg for arg in args.values()]
            groups.append(group)
        return groups or None

    @staticmethod
    def _build_output_type_schema(schema):
        if isinstance(schema, CMDClsSchemaBase):
            schema = schema.implement
            assert schema is not None

        if isinstance(schema, CMDArraySchemaBase):
            output = CMDArrayOutput()
        elif isinstance(schema, CMDObjectSchemaBase):
            output = CMDObjectOutput()
        elif isinstance(schema, CMDStringSchemaBase):
            output = CMDStringOutput()
        else:
            raise exceptions.InvalidAPIUsage(
                f"Invalid output schema, not support: {schema.type}"
            )
        return output

    @staticmethod
    def _build_output_type_by_subresource_selector(subresource_selector):
        if isinstance(subresource_selector, CMDJsonSubresourceSelector):
            index = subresource_selector.json
            pref_index = None
            while index:
                pref_index = index
                index = None
                if isinstance(pref_index, CMDObjectIndexBase):
                    if pref_index.prop:
                        index = pref_index.prop
                    elif pref_index.discriminator:
                        index = pref_index.discriminator
                    elif pref_index.additional_props:
                        if pref_index.additional_props.item:
                            index = pref_index.additional_props.item
                elif isinstance(pref_index, CMDObjectIndexDiscriminator):
                    if pref_index.prop:
                        index = pref_index.prop
                    elif pref_index.discriminator:
                        index = pref_index.discriminator
                elif isinstance(pref_index, CMDArrayIndexBase):
                    if pref_index.item:
                        index = pref_index.item
                else:
                    raise NotImplementedError()
            if isinstance(pref_index, (CMDObjectIndexBase, CMDObjectIndexDiscriminator)):
                output = CMDObjectOutput()
            elif isinstance(pref_index, CMDArrayIndexBase):
                output = CMDArrayOutput()
        else:
            raise NotImplementedError()
        return output


def handle_duplicated_options(arguments, has_subresource, operation_id):
    def can_be_replaced(arg1, arg2):
        # check whether you need to replace argument
        ret = False
        if _compare_argument(arg1, arg2, has_subresource):
            arg2.ref_schema.arg = arg1.var
            dropped_args.add(arg2.var)
            ret = False

        elif _compare_argument(arg2, arg1, has_subresource):
            arg1.ref_schema.arg = arg2.var
            dropped_args.add(arg1.var)
            ret = True

        else:
            # warning developer handle duplicated options
            logger.warning(
                f"Duplicated Option Value: {set(arg1.options).intersection(arg2.options)} : "
                f"{arg1.var} with {arg2.var} : {operation_id}"
            )

        return ret

    # check argument with duplicated option names
    dropped_args = set()
    used_args = set()
    for arg in arguments.values():
        used_args.add(arg.var)
        if arg.var in dropped_args or not arg.options:
            continue

        for v in arguments.values():
            if v.var in used_args or v.var in dropped_args or arg.var == v.var or not v.options:
                continue

            if not set(arg.options).isdisjoint(v.options):
                if can_be_replaced(arg, v):
                    break

    return [arg for var, arg in arguments.items() if var not in dropped_args]


def _compare_argument(arg, old_arg, has_subresource):
    arg_prefix = arg.var.split('.')[0]
    old_prefix = old_arg.var.split('.')[0]

    if old_prefix in (CMDArgBuildPrefix.Query, CMDArgBuildPrefix.Header):
        return True

    if old_prefix == CMDArgBuildPrefix.Path:
        return False

    if arg_prefix in (CMDArgBuildPrefix.Query, CMDArgBuildPrefix.Header):
        # only support path argument to replace
        return False

    elif arg_prefix == CMDArgBuildPrefix.Path:
        # path argument
        if has_subresource:
            return False

        arg_schema_required = arg.ref_schema.required
        arg_schema_name = arg.ref_schema.name
        try:
            # temporary assign required and name for diff
            arg.ref_schema.required = old_arg.ref_schema.required
            if old_arg.ref_schema.name.lower() == "name" and "name" in arg.options:
                arg.ref_schema.name = old_arg.ref_schema.name
            diff = arg.ref_schema.diff(old_arg.ref_schema, level=CMDDiffLevelEnum.Structure)
            if diff:
                return False
            return True
        finally:
            arg.ref_schema.name = arg_schema_name
            arg.ref_schema.required = arg_schema_required
    else:
        # body argument
        diff = arg.ref_schema.diff(old_arg.ref_schema, level=CMDDiffLevelEnum.Structure)
        if diff:
            return False
        return True



import typing
from schematics.types import StringType
from utils.plane import PlaneEnum
from utils.client import CloudEnum


class PlaneField(StringType):

    def __init__(self, *args, **kwargs):
        super().__init__(
            regex=r'^({})|({}:[a-z0-9_\-.]+)$'.format(PlaneEnum.Mgmt, PlaneEnum._Data),
            *args, **kwargs
        )


class CloudField(StringType):

    def __init__(self, *args, **kwargs):
         super().__init__(
            choices=CloudEnum.choices(),
            *args, **kwargs
        )


__all__ = ["PlaneField"]
